"""Tests covering the workflow listing code."""

import json
import os
import tempfile
import time
from datetime import datetime
from pathlib import Path
from unittest import TestCase, mock

import pytest
from rich.console import Console

import nf_core.pipelines.list


class TestList(TestCase):
    """Class for list tests"""

    def setUp(self) -> None:
        # create a temporary directory that can be used by the tests in this file
        tmp = Path(tempfile.TemporaryDirectory().name)
        self.tmp_nxf = tmp / "nxf"
        self.tmp_nxf_str = str(self.tmp_nxf)
        os.environ["NXF_ASSETS"] = self.tmp_nxf_str

    @mock.patch("subprocess.check_output")
    def test_working_listcall(self, mock_subprocess):
        """Test that listing pipelines works"""
        wf_table = nf_core.pipelines.list.list_workflows()
        console = Console(record=True)
        console.print(wf_table)
        output = console.export_text()
        assert "rnaseq" in output
        assert "exoseq" not in output

    @mock.patch("subprocess.check_output")
    def test_working_listcall_archived(self, mock_subprocess):
        """Test that listing pipelines works, showing archived pipelines"""
        wf_table = nf_core.pipelines.list.list_workflows(show_archived=True)
        console = Console(record=True)
        console.print(wf_table)
        output = console.export_text()
        assert "exoseq" in output

    @mock.patch("subprocess.check_output")
    def test_working_listcall_json(self, mock_subprocess):
        """Test that listing pipelines with JSON works"""
        wf_json_str = nf_core.pipelines.list.list_workflows(as_json=True)
        wf_json = json.loads(wf_json_str)
        for wf in wf_json["remote_workflows"]:
            if wf["name"] == "ampliseq":
                break
        else:
            raise AssertionError("Could not find ampliseq in JSON")

    def test_pretty_datetime(self):
        """Test that the pretty datetime function works"""
        now = datetime.now()
        nf_core.pipelines.list.pretty_date(now)
        now_ts = time.mktime(now.timetuple())
        nf_core.pipelines.list.pretty_date(now_ts)

    def test_local_workflows_and_fail(self):
        """Test the local workflow class and try to get local
        Nextflow workflow information"""
        loc_wf = nf_core.pipelines.list.LocalWorkflow("myWF")
        with pytest.raises(RuntimeError):
            loc_wf.get_local_nf_workflow_details()

    def test_local_workflows_compare_and_fail_silently(self):
        """Test the workflow class and try to compare local
        and remote workflows"""
        wfs = nf_core.pipelines.list.Workflows()
        lwf_ex = nf_core.pipelines.list.LocalWorkflow("myWF")
        lwf_ex.full_name = "my Workflow"
        lwf_ex.commit_sha = "aw3s0meh1sh"

        remote = {
            "name": "myWF",
            "full_name": "my Workflow",
            "description": "...",
            "archived": False,
            "stargazers_count": 42,
            "watchers_count": 6,
            "forks_count": 7,
            "releases": [],
        }

        rwf_ex = nf_core.pipelines.list.RemoteWorkflow(remote)
        rwf_ex.commit_sha = "aw3s0meh1sh"
        rwf_ex.releases = [{"tag_sha": "aw3s0meh1sh"}]

        wfs.local_workflows.append(lwf_ex)
        wfs.remote_workflows.append(rwf_ex)
        wfs.compare_remote_local()

        self.assertEqual(rwf_ex.local_wf, lwf_ex)

        rwf_ex.releases = []
        rwf_ex.releases.append({"tag_sha": "noaw3s0meh1sh"})
        wfs.compare_remote_local()

        rwf_ex.full_name = "your Workflow"
        wfs.compare_remote_local()

        rwf_ex.releases = None

    @mock.patch("nf_core.pipelines.list.LocalWorkflow")
    def test_parse_local_workflow_and_succeed(self, mock_local_wf):
        test_path = self.tmp_nxf / "nf-core"
        if not os.path.isdir(test_path):
            os.makedirs(test_path)
        assert os.environ["NXF_ASSETS"] == self.tmp_nxf_str
        with open(self.tmp_nxf / "nf-core/dummy-wf", "w") as f:
            f.write("dummy")
        workflows_obj = nf_core.pipelines.list.Workflows()
        workflows_obj.get_local_nf_workflows()
        assert len(workflows_obj.local_workflows) == 1

    @mock.patch("nf_core.pipelines.list.LocalWorkflow")
    @mock.patch("subprocess.check_output")
    def test_parse_local_workflow_home(self, mock_local_wf, mock_subprocess):
        test_path = self.tmp_nxf / "nf-core"
        if not os.path.isdir(test_path):
            os.makedirs(test_path)
        assert os.environ["NXF_ASSETS"] == self.tmp_nxf_str
        with open(self.tmp_nxf / "nf-core/dummy-wf", "w") as f:
            f.write("dummy")
        workflows_obj = nf_core.pipelines.list.Workflows()
        workflows_obj.get_local_nf_workflows()

    @mock.patch("os.stat")
    @mock.patch("git.Repo")
    def test_local_workflow_investigation(self, mock_repo, mock_stat):
        local_wf = nf_core.pipelines.list.LocalWorkflow("dummy")
        local_wf.local_path = self.tmp_nxf.parent
        mock_repo.head.commit.hexsha = "h00r4y"
        mock_stat.st_mode = 1
        local_wf.get_local_nf_workflow_details()

    def test_worflow_filter(self):
        workflows_obj = nf_core.pipelines.list.Workflows(["rna", "myWF"])

        remote = {
            "name": "myWF",
            "full_name": "my Workflow",
            "description": "rna",
            "archived": False,
            "stargazers_count": 42,
            "watchers_count": 6,
            "forks_count": 7,
            "releases": [],
        }

        rwf_ex = nf_core.pipelines.list.RemoteWorkflow(remote)
        rwf_ex.commit_sha = "aw3s0meh1sh"
        rwf_ex.releases = [{"tag_sha": "aw3s0meh1sh"}]

        remote2 = {
            "name": "myWF",
            "full_name": "my Workflow",
            "description": "dna",
            "archived": False,
            "stargazers_count": 42,
            "watchers_count": 6,
            "forks_count": 7,
            "releases": [],
        }

        rwf_ex2 = nf_core.pipelines.list.RemoteWorkflow(remote2)
        rwf_ex2.commit_sha = "aw3s0meh1sh"
        rwf_ex2.releases = [{"tag_sha": "aw3s0meh1sh"}]

        workflows_obj.remote_workflows.append(rwf_ex)
        workflows_obj.remote_workflows.append(rwf_ex2)

        assert len(workflows_obj.filtered_workflows()) == 1

    def test_filter_archived_workflows(self):
        """
        Test that archived workflows are not shown by default
        """
        workflows_obj = nf_core.pipelines.list.Workflows()
        remote1 = {"name": "myWF", "full_name": "my Workflow", "archived": True, "releases": []}
        rwf_ex1 = nf_core.pipelines.list.RemoteWorkflow(remote1)
        remote2 = {"name": "myWF", "full_name": "my Workflow", "archived": False, "releases": []}
        rwf_ex2 = nf_core.pipelines.list.RemoteWorkflow(remote2)

        workflows_obj.remote_workflows.append(rwf_ex1)
        workflows_obj.remote_workflows.append(rwf_ex2)

        filtered_workflows = workflows_obj.filtered_workflows()
        expected_workflows = [rwf_ex2]

        assert filtered_workflows == expected_workflows

    def test_show_archived_workflows(self):
        """
        Test that archived workflows can be shown optionally
        """
        workflows_obj = nf_core.pipelines.list.Workflows(show_archived=True)
        remote1 = {"name": "myWF", "full_name": "my Workflow", "archived": True, "releases": []}
        rwf_ex1 = nf_core.pipelines.list.RemoteWorkflow(remote1)
        remote2 = {"name": "myWF", "full_name": "my Workflow", "archived": False, "releases": []}
        rwf_ex2 = nf_core.pipelines.list.RemoteWorkflow(remote2)

        workflows_obj.remote_workflows.append(rwf_ex1)
        workflows_obj.remote_workflows.append(rwf_ex2)

        filtered_workflows = workflows_obj.filtered_workflows()
        expected_workflows = [rwf_ex1, rwf_ex2]

        assert filtered_workflows == expected_workflows



from pathlib import Path

import yaml

import nf_core.pipelines.lint

from ..test_lint import TestLint


class TestLintActionsCi(TestLint):
    def test_actions_ci_pass(self):
        """Lint test: actions_ci - PASS"""
        self.lint_obj._load()
        results = self.lint_obj.actions_ci()
        assert results["passed"] == [
            "'.github/workflows/ci.yml' is triggered on expected events",
            "'.github/workflows/ci.yml' checks minimum NF version",
        ]
        assert len(results.get("warned", [])) == 0
        assert len(results.get("failed", [])) == 0
        assert len(results.get("ignored", [])) == 0

    def test_actions_ci_fail_wrong_nf(self):
        """Lint test: actions_ci - FAIL - wrong minimum version of Nextflow tested"""
        self.lint_obj._load()
        self.lint_obj.minNextflowVersion = "1.2.3"
        results = self.lint_obj.actions_ci()
        assert results["failed"] == ["Minimum pipeline NF version '1.2.3' is not tested in '.github/workflows/ci.yml'"]

    def test_actions_ci_fail_wrong_trigger(self):
        """Lint test: actions_actions_ci - FAIL - workflow triggered incorrectly, NF ver not checked at all"""

        # Edit .github/workflows/actions_ci.yml to mess stuff up!
        new_pipeline = self._make_pipeline_copy()
        with open(Path(new_pipeline, ".github", "workflows", "ci.yml")) as fh:
            ci_yml = yaml.safe_load(fh)
        ci_yml[True]["push"] = ["dev", "patch"]
        ci_yml["jobs"]["test"]["strategy"]["matrix"] = {"nxf_versionnn": ["foo", ""]}
        with open(Path(new_pipeline, ".github", "workflows", "ci.yml"), "w") as fh:
            yaml.dump(ci_yml, fh)

        # Make lint object
        lint_obj = nf_core.pipelines.lint.PipelineLint(new_pipeline)
        lint_obj._load()

        results = lint_obj.actions_ci()
        assert results["failed"] == [
            "'.github/workflows/ci.yml' is not triggered on expected events",
            "'.github/workflows/ci.yml' does not check minimum NF version",
        ]



from pathlib import Path

import yaml

import nf_core.pipelines.create
import nf_core.pipelines.lint

from ..test_lint import TestLint


class TestLintConfigs(TestLint):
    def setUp(self) -> None:
        super().setUp()
        self.new_pipeline = self._make_pipeline_copy()

    def test_withname_in_modules_config(self):
        """Tests finding withName in modules.config passes linting."""

        lint_obj = nf_core.pipelines.lint.PipelineLint(self.new_pipeline)
        lint_obj._load()
        result = lint_obj.modules_config()
        assert len(result["failed"]) == 0
        assert any(
            ["`FASTQC` found in `conf/modules.config` and Nextflow scripts." in passed for passed in result["passed"]]
        )

    def test_superfluous_withname_in_modules_config_fails(self):
        """Tests finding withName in modules.config fails linting."""

        # Add withName to modules.config
        modules_config = Path(self.new_pipeline) / "conf" / "modules.config"
        with open(modules_config, "a") as f:
            f.write("\nwithName: 'BPIPE' {\n cache = false \n}")
        lint_obj = nf_core.pipelines.lint.PipelineLint(self.new_pipeline, hide_progress=False)
        lint_obj._load()
        result = lint_obj.modules_config()
        assert len(result["failed"]) == 1
        assert result["failed"][0].startswith("`conf/modules.config` contains `withName:BPIPE`")

    def test_ignore_modules_config(self):
        """Tests ignoring the modules.config passes linting."""

        # ignore modules.config in linting
        with open(Path(self.new_pipeline) / ".nf-core.yml") as f:
            content = yaml.safe_load(f)
            old_content = content.copy()
            content["lint"] = {"modules_config": False}
        with open(Path(self.new_pipeline) / ".nf-core.yml", "w") as f:
            yaml.dump(content, f)
        Path(self.new_pipeline, "conf", "modules.config").unlink()
        lint_obj = nf_core.pipelines.lint.PipelineLint(self.new_pipeline)
        lint_obj._load()
        result = lint_obj.modules_config()
        assert len(result["ignored"]) == 1
        assert result["ignored"][0].startswith("`conf/modules.config` not found, but it is ignored.")
        # cleanup
        with open(Path(self.new_pipeline) / ".nf-core.yml", "w") as f:
            yaml.dump(old_content, f)

    def test_superfluous_withname_in_base_config_fails(self):
        """Tests finding withName in base.config fails linting."""

        # Add withName to base.config
        base_config = Path(self.new_pipeline) / "conf" / "base.config"
        with open(base_config, "a") as f:
            f.write("\nwithName:CUSTOM_DUMPSOFTWAREVERSIONS {\n cache = false \n}")
        lint_obj = nf_core.pipelines.lint.PipelineLint(self.new_pipeline)
        lint_obj._load()
        result = lint_obj.base_config()
        assert len(result["failed"]) == 1
        assert result["failed"][0].startswith("`conf/base.config` contains `withName:CUSTOM_DUMPSOFTWAREVERSIONS`")

    def test_ignore_base_config(self):
        """Tests ignoring the base.config passes linting."""

        # ignore base.config in linting
        with open(Path(self.new_pipeline) / ".nf-core.yml") as f:
            content = yaml.safe_load(f)
            old_content = content.copy()
            content["lint"] = {"base_config": False}
        with open(Path(self.new_pipeline) / ".nf-core.yml", "w") as f:
            yaml.dump(content, f)
        Path(self.new_pipeline, "conf", "base.config").unlink()
        lint_obj = nf_core.pipelines.lint.PipelineLint(self.new_pipeline)
        lint_obj._load()
        result = lint_obj.base_config()
        assert len(result["ignored"]) == 1
        assert result["ignored"][0].startswith("`conf/base.config` not found, but it is ignored.")
        # cleanup
        with open(Path(self.new_pipeline) / ".nf-core.yml", "w") as f:
            yaml.dump(old_content, f)



import re
from contextlib import contextmanager
from contextvars import ContextVar
from logging import LogRecord
from pathlib import Path
from typing import Any, Dict, Iterator, Union

import yaml
from pydantic import ConfigDict, ValidationError, ValidationInfo, field_validator
from rich.logging import RichHandler
from textual import on
from textual._context import active_app
from textual.app import ComposeResult
from textual.containers import Grid, HorizontalScroll
from textual.message import Message
from textual.validation import ValidationResult, Validator
from textual.widget import Widget
from textual.widgets import Button, Input, Markdown, RichLog, Static, Switch

import nf_core
from nf_core.utils import NFCoreTemplateConfig

# Use ContextVar to define a context on the model initialization
_init_context_var: ContextVar = ContextVar("_init_context_var", default={})


@contextmanager
def init_context(value: Dict[str, Any]) -> Iterator[None]:
    token = _init_context_var.set(value)
    try:
        yield
    finally:
        _init_context_var.reset(token)


# Define a global variable to store the pipeline type
NFCORE_PIPELINE_GLOBAL: bool = True

# YAML file describing template features
features_yml_path = Path(nf_core.__file__).parent / "pipelines" / "create" / "template_features.yml"


class CreateConfig(NFCoreTemplateConfig):
    """Pydantic model for the nf-core create config."""

    model_config = ConfigDict(extra="allow")

    def __init__(self, /, **data: Any) -> None:
        """Custom init method to allow using a context on the model initialization."""
        self.__pydantic_validator__.validate_python(
            data,
            self_instance=self,
            context=_init_context_var.get(),
        )

    @field_validator("name")
    @classmethod
    def name_nospecialchars(cls, v: str, info: ValidationInfo) -> str:
        """Check that the pipeline name is simple."""
        context = info.context
        if context and context["is_nfcore"]:
            if not re.match(r"^[a-z]+$", v):
                raise ValueError("Must be lowercase without punctuation.")
        else:
            if not re.match(r"^[-\w]+$", v):
                raise ValueError("Must not contain special characters. Only '-' or '_' are allowed.")
        return v

    @field_validator("org", "description", "author", "version", "outdir")
    @classmethod
    def notempty(cls, v: str) -> str:
        """Check that string values are not empty."""
        if v.strip() == "":
            raise ValueError("Cannot be left empty.")
        return v

    @field_validator("version")
    @classmethod
    def version_nospecialchars(cls, v: str) -> str:
        """Check that the pipeline version is simple."""
        if not re.match(r"^([0-9]+)(\.?([0-9]+))*(dev)?$", v):
            raise ValueError(
                "Must contain at least one number, and can be prefixed by 'dev'. Do not use a 'v' prefix or spaces."
            )
        return v

    @field_validator("outdir")
    @classmethod
    def path_valid(cls, v: str) -> str:
        """Check that a path is valid."""
        if not Path(v).is_dir():
            raise ValueError("Must be a valid path.")
        return v


class TextInput(Static):
    """Widget for text inputs.

    Provides standard interface for a text input with help text
    and validation messages.
    """

    def __init__(self, field_id, placeholder, description, default="", password=False, **kwargs) -> None:
        """Initialise the widget with our values.

        Pass on kwargs upstream for standard usage."""
        super().__init__(**kwargs)
        self.field_id: str = field_id
        self.id: str = field_id
        self.placeholder: str = placeholder
        self.description: str = description
        self.default: str = default
        self.password: bool = password

    def compose(self) -> ComposeResult:
        yield Grid(
            Static(self.description, classes="field_help"),
            Input(
                placeholder=self.placeholder,
                validators=[ValidateConfig(self.field_id)],
                value=self.default,
                password=self.password,
            ),
            Static(classes="validation_msg"),
            classes="text-input-grid",
        )

    @on(Input.Changed)
    @on(Input.Submitted)
    def show_invalid_reasons(self, event: Union[Input.Changed, Input.Submitted]) -> None:
        """Validate the text input and show errors if invalid."""
        val_msg = self.query_one(".validation_msg")
        if not isinstance(val_msg, Static):
            raise ValueError("Validation message not found.")

        if event.validation_result is not None and not event.validation_result.is_valid:
            # check that val_msg is instance of Static
            if isinstance(val_msg, Static):
                val_msg.update("\n".join(event.validation_result.failure_descriptions))
        else:
            val_msg.update("")


class ValidateConfig(Validator):
    """Validate any config value, using Pydantic."""

    def __init__(self, key) -> None:
        """Initialise the validator with the model key to validate."""
        super().__init__()
        self.key = key

    def validate(self, value: str) -> ValidationResult:
        """Try creating a Pydantic object with this key set to this value.

        If it fails, return the error messages."""
        try:
            with init_context({"is_nfcore": NFCORE_PIPELINE_GLOBAL}):
                CreateConfig(**{f"{self.key}": value})
                return self.success()
        except ValidationError as e:
            return self.failure(", ".join([err["msg"] for err in e.errors()]))


class HelpText(Markdown):
    """A class to show a text box with help text."""

    def __init__(self, **kwargs) -> None:
        super().__init__(**kwargs)

    def show(self) -> None:
        """Method to show the help text box."""
        self.add_class("displayed")

    def hide(self) -> None:
        """Method to hide the help text box."""
        self.remove_class("displayed")


class PipelineFeature(Static):
    """Widget for the selection of pipeline features."""

    def __init__(self, markdown: str, title: str, subtitle: str, field_id: str, **kwargs) -> None:
        super().__init__(**kwargs)
        self.markdown = markdown
        self.title = title
        self.subtitle = subtitle
        self.field_id = field_id

    def on_button_pressed(self, event: Button.Pressed) -> None:
        """When the button is pressed, change the type of the button."""
        if event.button.id == "show_help":
            self.add_class("displayed")
        elif event.button.id == "hide_help":
            self.remove_class("displayed")

    def compose(self) -> ComposeResult:
        """
        Create child widgets.

        Displayed row with a switch, a short text description and a help button.
        Hidden row with a help text box.
        """
        yield HorizontalScroll(
            Switch(value=True, id=self.field_id),
            Static(self.title, classes="feature_title"),
            Static(self.subtitle, classes="feature_subtitle"),
            Button("Show help", id="show_help", variant="primary"),
            Button("Hide help", id="hide_help"),
            classes="custom_grid",
        )
        yield HelpText(markdown=self.markdown, classes="help_box")


class LoggingConsole(RichLog):
    file = False
    console: Widget

    def print(self, content):
        self.write(content)


class CustomLogHandler(RichHandler):
    """A Logging handler which extends RichHandler to write to a Widget and handle a Textual App."""

    def emit(self, record: LogRecord) -> None:
        """Invoked by logging."""
        try:
            _app = active_app.get()
        except LookupError:
            pass
        else:
            super().emit(record)


class ShowLogs(Message):
    """Custom message to show the logging messages."""

    pass


## Functions
def add_hide_class(app, widget_id: str) -> None:
    """Add class 'hide' to a widget. Not display widget."""
    app.get_widget_by_id(widget_id).add_class("hide")


def remove_hide_class(app, widget_id: str) -> None:
    """Remove class 'hide' to a widget. Display widget."""
    app.get_widget_by_id(widget_id).remove_class("hide")


def load_features_yaml() -> Dict:
    """Load the YAML file describing template features."""
    with open(features_yml_path) as fh:
        return yaml.safe_load(fh)



import json
import shutil
from pathlib import Path

import nf_core.subworkflows

from ..test_subworkflows import TestSubworkflows
from ..utils import GITLAB_SUBWORKFLOWS_BRANCH, GITLAB_URL


class TestSubworkflowsLint(TestSubworkflows):
    def test_subworkflows_lint(self):
        """Test linting the fastq_align_bowtie2 subworkflow"""
        self.subworkflow_install.install("fastq_align_bowtie2")
        subworkflow_lint = nf_core.subworkflows.SubworkflowLint(directory=self.pipeline_dir)
        subworkflow_lint.lint(print_results=False, subworkflow="fastq_align_bowtie2")
        assert len(subworkflow_lint.failed) == 0, f"Linting failed with {[x.__dict__ for x in subworkflow_lint.failed]}"
        assert len(subworkflow_lint.passed) > 0
        assert len(subworkflow_lint.warned) >= 0

    def test_subworkflows_lint_empty(self):
        """Test linting a pipeline with no subworkflows installed"""
        self.subworkflow_remove.remove("utils_nextflow_pipeline", force=True)
        self.subworkflow_remove.remove("utils_nfcore_pipeline", force=True)
        self.subworkflow_remove.remove("utils_nfschema_plugin", force=True)
        nf_core.subworkflows.SubworkflowLint(directory=self.pipeline_dir)
        assert "No subworkflows from https://github.com/nf-core/modules.git installed in pipeline" in self.caplog.text

    def test_subworkflows_lint_new_subworkflow(self):
        """lint a new subworkflow"""
        subworkflow_lint = nf_core.subworkflows.SubworkflowLint(directory=self.nfcore_modules)
        subworkflow_lint.lint(print_results=True, all_subworkflows=True)
        assert len(subworkflow_lint.failed) == 0

        assert len(subworkflow_lint.passed) > 0
        assert len(subworkflow_lint.warned) >= 0

    def test_subworkflows_lint_no_gitlab(self):
        """Test linting a pipeline with no subworkflows installed"""
        nf_core.subworkflows.SubworkflowLint(directory=self.pipeline_dir, remote_url=GITLAB_URL)
        assert f"No subworkflows from {GITLAB_URL} installed in pipeline" in self.caplog.text

    def test_subworkflows_lint_gitlab_subworkflows(self):
        """Lint subworkflows from a different remote"""
        self.subworkflow_install_gitlab.install("bam_stats_samtools")
        subworkflow_lint = nf_core.subworkflows.SubworkflowLint(
            directory=self.pipeline_dir, remote_url=GITLAB_URL, branch=GITLAB_SUBWORKFLOWS_BRANCH
        )
        subworkflow_lint.lint(print_results=False, all_subworkflows=True)
        assert len(subworkflow_lint.failed) == 0
        assert len(subworkflow_lint.passed) > 0
        assert len(subworkflow_lint.warned) >= 0

    def test_subworkflows_lint_multiple_remotes(self):
        """Lint subworkflows from a different remote"""
        self.subworkflow_install_gitlab.install("bam_stats_samtools")
        self.subworkflow_install.install("fastq_align_bowtie2")
        subworkflow_lint = nf_core.subworkflows.SubworkflowLint(
            directory=self.pipeline_dir, remote_url=GITLAB_URL, branch=GITLAB_SUBWORKFLOWS_BRANCH
        )
        subworkflow_lint.lint(print_results=False, all_subworkflows=True)
        assert len(subworkflow_lint.failed) == 0
        assert len(subworkflow_lint.passed) > 0
        assert len(subworkflow_lint.warned) >= 0

    def test_subworkflows_lint_update_meta_yml(self):
        """update the meta.yml of a subworkflow"""
        subworkflow_lint = nf_core.subworkflows.SubworkflowLint(directory=self.nfcore_modules, fix=True)
        subworkflow_lint.lint(print_results=False, subworkflow="test_subworkflow")
        assert len(subworkflow_lint.failed) == 0, f"Linting failed with {[x.__dict__ for x in subworkflow_lint.failed]}"
        assert len(subworkflow_lint.passed) > 0
        assert len(subworkflow_lint.warned) >= 0

    def test_subworkflows_lint_snapshot_file(self):
        """Test linting a subworkflow with a snapshot file"""
        subworkflow_lint = nf_core.subworkflows.SubworkflowLint(directory=self.nfcore_modules)
        subworkflow_lint.lint(print_results=False, subworkflow="test_subworkflow")
        assert len(subworkflow_lint.failed) == 0, f"Linting failed with {[x.__dict__ for x in subworkflow_lint.failed]}"
        assert len(subworkflow_lint.passed) > 0
        assert len(subworkflow_lint.warned) >= 0

    def test_subworkflows_lint_snapshot_file_missing_fail(self):
        """Test linting a subworkflow with a snapshot file missing, which should fail"""
        Path(
            self.nfcore_modules,
            "subworkflows",
            "nf-core",
            "test_subworkflow",
            "tests",
            "main.nf.test.snap",
        ).unlink()
        subworkflow_lint = nf_core.subworkflows.SubworkflowLint(directory=self.nfcore_modules)
        subworkflow_lint.lint(print_results=False, subworkflow="test_subworkflow")
        Path(
            self.nfcore_modules,
            "subworkflows",
            "nf-core",
            "test_subworkflow",
            "tests",
            "main.nf.test.snap",
        ).touch()
        assert len(subworkflow_lint.failed) == 1, f"Linting failed with {[x.__dict__ for x in subworkflow_lint.failed]}"
        assert len(subworkflow_lint.passed) > 0
        assert len(subworkflow_lint.warned) >= 0

    def test_subworkflows_lint_snapshot_file_not_needed(self):
        """Test linting a subworkflow which doesn't need a snapshot file by removing the snapshot keyword in the main.nf.test file"""
        with open(
            Path(
                self.nfcore_modules,
                "subworkflows",
                "nf-core",
                "test_subworkflow",
                "tests",
                "main.nf.test",
            )
        ) as fh:
            content = fh.read()
            new_content = content.replace("snapshot(", "snap (")
        with open(
            Path(
                self.nfcore_modules,
                "subworkflows",
                "nf-core",
                "test_subworkflow",
                "tests",
                "main.nf.test",
            ),
            "w",
        ) as fh:
            fh.write(new_content)

        Path(
            self.nfcore_modules,
            "subworkflows",
            "nf-core",
            "test_subworkflow",
            "tests",
            "main.nf.test.snap",
        ).unlink()
        subworkflow_lint = nf_core.subworkflows.SubworkflowLint(directory=self.nfcore_modules)
        subworkflow_lint.lint(print_results=False, subworkflow="test_subworkflow")
        Path(
            self.nfcore_modules,
            "subworkflows",
            "nf-core",
            "test_subworkflow",
            "tests",
            "main.nf.test.snap",
        ).touch()
        assert len(subworkflow_lint.failed) == 0, f"Linting failed with {[x.__dict__ for x in subworkflow_lint.failed]}"
        assert len(subworkflow_lint.passed) > 0
        assert len(subworkflow_lint.warned) >= 0

    def test_subworkflows_lint_less_than_two_modules_warning(self):
        """Test linting a subworkflow with less than two modules"""
        self.subworkflow_install.install("bam_stats_samtools")
        # Remove two modules
        with open(
            Path(
                self.pipeline_dir,
                "subworkflows",
                "nf-core",
                "bam_stats_samtools",
                "main.nf",
            )
        ) as fh:
            content = fh.read()
            new_content = content.replace(
                "include { SAMTOOLS_IDXSTATS } from '../../../modules/nf-core/samtools/idxstats/main'",
                "",
            )
            new_content = new_content.replace(
                "include { SAMTOOLS_FLAGSTAT } from '../../../modules/nf-core/samtools/flagstat/main'",
                "",
            )
        with open(
            Path(
                self.pipeline_dir,
                "subworkflows",
                "nf-core",
                "bam_stats_samtools",
                "main.nf",
            ),
            "w",
        ) as fh:
            fh.write(new_content)
        subworkflow_lint = nf_core.subworkflows.SubworkflowLint(directory=self.pipeline_dir)
        subworkflow_lint.lint(print_results=False, subworkflow="bam_stats_samtools")
        assert len(subworkflow_lint.failed) >= 0, f"Linting failed with {[x.__dict__ for x in subworkflow_lint.failed]}"
        assert len(subworkflow_lint.passed) > 0
        assert len(subworkflow_lint.warned) > 0
        assert subworkflow_lint.warned[0].lint_test == "main_nf_include"
        # cleanup
        self.subworkflow_remove.remove("bam_stats_samtools", force=True)

    def test_subworkflows_lint_include_multiple_alias(self):
        """Test linting a subworkflow with multiple include methods"""
        self.subworkflow_install.install("bam_stats_samtools")
        with open(
            Path(
                self.pipeline_dir,
                "subworkflows",
                "nf-core",
                "bam_stats_samtools",
                "main.nf",
            )
        ) as fh:
            content = fh.read()
            new_content = content.replace("SAMTOOLS_STATS", "SAMTOOLS_STATS_1")
            new_content = new_content.replace(
                "include { SAMTOOLS_STATS_1 ",
                "include { SAMTOOLS_STATS as SAMTOOLS_STATS_1; SAMTOOLS_STATS as SAMTOOLS_STATS_2 ",
            )
        with open(
            Path(
                self.pipeline_dir,
                "subworkflows",
                "nf-core",
                "bam_stats_samtools",
                "main.nf",
            ),
            "w",
        ) as fh:
            fh.write(new_content)

        subworkflow_lint = nf_core.subworkflows.SubworkflowLint(directory=self.pipeline_dir)
        subworkflow_lint.lint(print_results=False, subworkflow="bam_stats_samtools")
        assert len(subworkflow_lint.failed) >= 0, f"Linting failed with {[x.__dict__ for x in subworkflow_lint.failed]}"
        assert len(subworkflow_lint.passed) > 0
        assert len(subworkflow_lint.warned) == 2
        assert any(
            [
                x.message == "Included component 'SAMTOOLS_STATS_1' versions are added in main.nf"
                for x in subworkflow_lint.passed
            ]
        )
        assert any(
            [x.message == "Included component 'SAMTOOLS_STATS_1' used in main.nf" for x in subworkflow_lint.passed]
        )
        assert any(
            [x.message == "Included component 'SAMTOOLS_STATS_2' not used in main.nf" for x in subworkflow_lint.warned]
        )

        # cleanup
        self.subworkflow_remove.remove("bam_stats_samtools", force=True)

    def test_subworkflows_lint_capitalization_fail(self):
        """Test linting a subworkflow with a capitalization fail"""
        self.subworkflow_install.install("bam_stats_samtools")
        # change workflow name to lowercase
        with open(
            Path(
                self.pipeline_dir,
                "subworkflows",
                "nf-core",
                "bam_stats_samtools",
                "main.nf",
            )
        ) as fh:
            content = fh.read()
            new_content = content.replace("workflow BAM_STATS_SAMTOOLS {", "workflow bam_stats_samtools {")
        with open(
            Path(
                self.pipeline_dir,
                "subworkflows",
                "nf-core",
                "bam_stats_samtools",
                "main.nf",
            ),
            "w",
        ) as fh:
            fh.write(new_content)
        subworkflow_lint = nf_core.subworkflows.SubworkflowLint(directory=self.pipeline_dir)
        subworkflow_lint.lint(print_results=False, subworkflow="bam_stats_samtools")
        assert len(subworkflow_lint.failed) >= 1, f"Linting failed with {[x.__dict__ for x in subworkflow_lint.failed]}"
        assert len(subworkflow_lint.passed) > 0
        assert len(subworkflow_lint.warned) >= 0
        assert any([x.lint_test == "workflow_capitals" for x in subworkflow_lint.failed])

        # cleanup
        self.subworkflow_remove.remove("bam_stats_samtools", force=True)

    def test_subworkflows_absent_version(self):
        """Test linting a nf-test subworkflow if the versions is absent in the snapshot file `"""
        snap_file = Path(
            self.nfcore_modules,
            "subworkflows",
            "nf-core",
            "test_subworkflow",
            "tests",
            "main.nf.test.snap",
        )
        with open(snap_file) as fh:
            content = fh.read()
            new_content = content.replace("versions", "foo")
        with open(snap_file, "w") as fh:
            fh.write(new_content)

        subworkflow_lint = nf_core.subworkflows.SubworkflowLint(directory=self.nfcore_modules)
        subworkflow_lint.lint(print_results=False, subworkflow="test_subworkflow")
        assert len(subworkflow_lint.failed) == 0
        assert len(subworkflow_lint.passed) > 0
        assert len(subworkflow_lint.warned) >= 0, f"Linting warned with {[x.__dict__ for x in subworkflow_lint.warned]}"
        assert any([x.lint_test == "test_snap_versions" for x in subworkflow_lint.warned])

        # cleanup
        with open(snap_file, "w") as fh:
            fh.write(content)

    def test_subworkflows_missing_test_dir(self):
        """Test linting a nf-test subworkflow if the tests directory is missing"""
        test_dir = Path(self.nfcore_modules, "subworkflows", "nf-core", "test_subworkflow", "tests")
        test_dir_copy = shutil.copytree(test_dir, test_dir.parent / "tests_copy")
        shutil.rmtree(test_dir)

        subworkflow_lint = nf_core.subworkflows.SubworkflowLint(self.nfcore_modules)
        subworkflow_lint.lint(print_results=False, subworkflow="test_subworkflow")
        assert len(subworkflow_lint.failed) == 1
        assert len(subworkflow_lint.passed) > 0
        assert len(subworkflow_lint.warned) >= 0, f"Linting warned with {[x.__dict__ for x in subworkflow_lint.warned]}"
        assert any([x.lint_test == "test_dir_exists" for x in subworkflow_lint.failed])

        # cleanup
        shutil.copytree(test_dir_copy, test_dir)

    # There are many steps before the actual main_nf linting where we rely on the main_nf file to exist, so this test is not possible for now
    # def test_subworkflows_missing_main_nf(self):
    #     """Test linting a nf-test subworkflow if the main.nf file is missing"""

    #     subworkflow_lint = nf_core.subworkflows.SubworkflowLint(directory=self.nfcore_modules)
    #     main_nf = Path(self.nfcore_modules, "subworkflows", "nf-core", "test_subworkflow", "main.nf")
    #     main_nf_copy = shutil.copy(main_nf, main_nf.parent / "main_nf_copy")
    #     main_nf.unlink()
    #     subworkflow_lint.lint(print_results=False, subworkflow="test_subworkflow")
    #     assert len(subworkflow_lint.failed) == 1, f"Linting failed with {[x.__dict__ for x in subworkflow_lint.failed]}"
    #     assert len(subworkflow_lint.passed) > 0
    #     assert len(subworkflow_lint.warned) >= 0
    #     assert subworkflow_lint.failed[0].lint_test == "main_nf_exists"

    #     # cleanup
    #     shutil.copy(main_nf_copy, main_nf)
    #     shutil.rmtree(Path(self.nfcore_modules, "subworkflows", "nf-core", "test_subworkflow_backup"))

    def test_subworkflows_empty_file_in_snapshot(self):
        """Test linting a nf-test subworkflow with an empty file sha sum in the test snapshot, which should make it fail (if it is not a stub)"""
        snap_file = Path(
            self.nfcore_modules,
            "subworkflows",
            "nf-core",
            "test_subworkflow",
            "tests",
            "main.nf.test.snap",
        )
        snap = json.load(snap_file.open())
        content = snap_file.read_text()
        snap["my test"]["content"][0]["0"] = "test:md5,d41d8cd98f00b204e9800998ecf8427e"

        with open(snap_file, "w") as fh:
            json.dump(snap, fh)

        subworkflow_lint = nf_core.subworkflows.SubworkflowLint(directory=self.nfcore_modules)
        subworkflow_lint.lint(print_results=False, subworkflow="test_subworkflow")
        assert len(subworkflow_lint.failed) == 1, f"Linting failed with {[x.__dict__ for x in subworkflow_lint.failed]}"
        assert len(subworkflow_lint.passed) > 0
        assert len(subworkflow_lint.warned) >= 0
        assert subworkflow_lint.failed[0].lint_test == "test_snap_md5sum"

        # reset the file
        with open(snap_file, "w") as fh:
            fh.write(content)

    def test_subworkflows_empty_file_in_stub_snapshot(self):
        """Test linting a nf-test subworkflow with an empty file sha sum in the stub test snapshot, which should make it not fail"""
        snap_file = Path(
            self.nfcore_modules,
            "subworkflows",
            "nf-core",
            "test_subworkflow",
            "tests",
            "main.nf.test.snap",
        )
        snap = json.load(snap_file.open())
        content = snap_file.read_text()
        snap["my_test_stub"] = {"content": [{"0": "test:md5,d41d8cd98f00b204e9800998ecf8427e", "versions": {}}]}

        with open(snap_file, "w") as fh:
            json.dump(snap, fh)

        subworkflow_lint = nf_core.subworkflows.SubworkflowLint(directory=self.nfcore_modules)
        subworkflow_lint.lint(print_results=False, subworkflow="test_subworkflow")
        assert len(subworkflow_lint.failed) == 0, f"Linting failed with {[x.__dict__ for x in subworkflow_lint.failed]}"
        assert len(subworkflow_lint.passed) > 0
        assert len(subworkflow_lint.warned) >= 0
        assert any(x.lint_test == "test_snap_md5sum" for x in subworkflow_lint.passed)

        # reset the file
        with open(snap_file, "w") as fh:
            fh.write(content)



import logging

from nf_core.components.create import ComponentCreate

log = logging.getLogger(__name__)


class ModuleCreate(ComponentCreate):
    def __init__(
        self,
        pipeline_dir,
        component="",
        author=None,
        process_label=None,
        has_meta=None,
        force=False,
        conda_name=None,
        conda_version=None,
        empty_template=False,
        migrate_pytest=False,
    ):
        super().__init__(
            "modules",
            pipeline_dir,
            component,
            author,
            process_label,
            has_meta,
            force,
            conda_name,
            conda_version,
            empty_template,
            migrate_pytest,
        )



from pathlib import Path

import pytest

from nf_core.modules.install import ModuleInstall
from nf_core.modules.modules_json import ModulesJson

from ..test_modules import TestModules
from ..utils import (
    GITLAB_BRANCH_ORG_PATH_BRANCH,
    GITLAB_BRANCH_TEST_BRANCH,
    GITLAB_REPO,
    GITLAB_URL,
    with_temporary_folder,
)


class TestModulesCreate(TestModules):
    @with_temporary_folder
    def test_modules_install_emptypipeline(self, tmpdir):
        """Test installing a module - empty dir given"""
        Path(tmpdir, "nf-core-pipe").mkdir()
        self.mods_install.directory = Path(tmpdir, "nf-core-pipe")
        with pytest.raises(UserWarning) as excinfo:
            self.mods_install.install("fastp")
        assert "Could not find a 'main.nf' or 'nextflow.config' file" in str(excinfo.value)

    def test_modules_install_nomodule(self):
        """Test installing a module - unrecognised module given"""
        with pytest.raises(ValueError) as excinfo:
            self.mods_install.install("foo")
        assert excinfo.typename == "ValueError"
        assert "Module 'foo' not found in available modules" in self.caplog.text

    def test_modules_install_trimgalore(self):
        """Test installing a module - TrimGalore!"""
        assert self.mods_install.install("trimgalore") is not False
        assert self.mods_install.directory is not None
        module_path = Path(self.mods_install.directory, "modules", "nf-core", "trimgalore")
        assert module_path.exists()

    def test_modules_install_trimgalore_twice(self):
        """Test installing a module - TrimGalore! already there"""
        self.mods_install.install("trimgalore")
        assert self.mods_install.install("trimgalore") is True

    def test_modules_install_from_gitlab(self):
        """Test installing a module from GitLab"""
        assert self.mods_install_gitlab.install("fastqc") is True

    def test_modules_install_different_branch_fail(self):
        """Test installing a module from a different branch"""
        install_obj = ModuleInstall(self.pipeline_dir, remote_url=GITLAB_URL, branch=GITLAB_BRANCH_TEST_BRANCH)
        # The FastQC module does not exists in the branch-test branch
        with pytest.raises(Exception) as excinfo:
            install_obj.install("fastqc")
            assert "Module 'fastqc' not found in available module" in str(excinfo.value)

    def test_modules_install_different_branch_succeed(self):
        """Test installing a module from a different branch"""
        install_obj = ModuleInstall(self.pipeline_dir, remote_url=GITLAB_URL, branch=GITLAB_BRANCH_TEST_BRANCH)
        # The fastp module does exists in the branch-test branch
        assert install_obj.install("fastp") is True

        # Verify that the branch entry was added correctly
        modules_json = ModulesJson(self.pipeline_dir)
        assert (
            modules_json.get_component_branch(self.component_type, "fastp", GITLAB_URL, GITLAB_REPO)
            == GITLAB_BRANCH_TEST_BRANCH
        )

    def test_modules_install_tracking(self):
        """Test installing a module and finding 'modules' in the installed_by section of modules.json"""
        self.mods_install.install("trimgalore")

        # Verify that the installed_by entry was added correctly
        modules_json = ModulesJson(self.pipeline_dir)
        mod_json = modules_json.get_modules_json()
        assert mod_json["repos"]["https://github.com/nf-core/modules.git"]["modules"]["nf-core"]["trimgalore"][
            "installed_by"
        ] == ["modules"]

    def test_modules_install_alternate_remote(self):
        """Test installing a module from a different remote with the same organization path"""
        install_obj = ModuleInstall(self.pipeline_dir, remote_url=GITLAB_URL, branch=GITLAB_BRANCH_ORG_PATH_BRANCH)
        # Install fastqc from GitLab which is also installed from GitHub with the same org_path
        with pytest.raises(Exception) as excinfo:
            install_obj.install("fastqc")
            assert "Could not find a 'main.nf' or 'nextflow.config' file" in str(excinfo.value)



import logging
import os
import shutil
import tempfile
from pathlib import Path

import questionary

import nf_core.modules.modules_utils
import nf_core.utils
from nf_core.components.components_command import ComponentCommand
from nf_core.components.components_differ import ComponentsDiffer
from nf_core.components.components_utils import (
    get_components_to_install,
    prompt_component_version_sha,
)
from nf_core.components.install import ComponentInstall
from nf_core.components.remove import ComponentRemove
from nf_core.modules.modules_json import ModulesJson
from nf_core.modules.modules_repo import ModulesRepo
from nf_core.utils import plural_es, plural_s, plural_y

log = logging.getLogger(__name__)


class ComponentUpdate(ComponentCommand):
    def __init__(
        self,
        pipeline_dir,
        component_type,
        force=False,
        prompt=False,
        sha=None,
        update_all=False,
        show_diff=None,
        save_diff_fn=None,
        update_deps=False,
        remote_url=None,
        branch=None,
        no_pull=False,
        limit_output=False,
    ):
        super().__init__(component_type, pipeline_dir, remote_url, branch, no_pull)
        self.force = force
        self.prompt = prompt
        self.sha = sha
        self.update_all = update_all
        self.show_diff = show_diff
        self.save_diff_fn = save_diff_fn
        self.limit_output = limit_output
        self.update_deps = update_deps
        self.component = None
        self.update_config = None
        self.modules_json = ModulesJson(self.directory)
        self.branch = branch

    def _parameter_checks(self):
        """Checks the compatibility of the supplied parameters.

        Raises:
            UserWarning: if any checks fail.
        """

        if self.save_diff_fn and self.show_diff:
            raise UserWarning("Either `--preview` or `--save_diff` can be specified, not both.")

        if self.update_all and self.component:
            raise UserWarning(f"Either a {self.component_type[:-1]} or the '--all' flag can be specified, not both.")

        if self.repo_type == "modules":
            raise UserWarning(
                f"{self.component_type.title()} can not be updated in clones of the nf-core/modules repository."
            )

        if self.prompt and self.sha is not None:
            raise UserWarning("Cannot use '--sha' and '--prompt' at the same time.")

        if not self.has_valid_directory():
            raise UserWarning("The command was not run in a valid pipeline directory.")
        if self.limit_output and not (self.save_diff_fn or self.show_diff):
            raise UserWarning("The '--limit-output' flag can only be used with '--preview' or '--save-diff'.")

    def update(self, component=None, silent=False, updated=None, check_diff_exist=True) -> bool:
        """Updates a specified module/subworkflow or all modules/subworkflows in a pipeline.

        If updating a subworkflow: updates all modules used in that subworkflow.
        If updating a module: updates all subworkflows that use the module.

        Args:
            component (str): The name of the module/subworkflow to update.

        Returns:
            (bool): True if the update was successful, False otherwise.
        """
        self.component = component
        if updated is None:
            updated = []

        _, tool_config = nf_core.utils.load_tools_config(self.directory)
        self.update_config = getattr(tool_config, "update", {}) or {}
        self._parameter_checks()

        # Check modules directory structure
        self.check_modules_structure()

        # Verify that 'modules.json' is consistent with the installed modules
        if not silent:
            self.modules_json.check_up_to_date()

        if not self.update_all and component is None:
            choices = [f"All {self.component_type}", f"Named {self.component_type[:-1]}"]
            self.update_all = (
                questionary.select(
                    f"Update all {self.component_type} or a single named {self.component_type[:-1]}?",
                    choices=choices,
                    style=nf_core.utils.nfcore_question_style,
                ).unsafe_ask()
                == f"All {self.component_type}"
            )

        # Verify that the provided SHA exists in the repo
        if self.sha is not None and not self.modules_repo.sha_exists_on_branch(self.sha):
            log.error(f"Commit SHA '{self.sha}' doesn't exist in '{self.modules_repo.remote_url}'")
            return False

        # Get the list of modules/subworkflows to update, and their version information
        components_info = (
            self.get_all_components_info() if self.update_all else [self.get_single_component_info(component)]
        )
        # Save the current state of the modules.json
        old_modules_json = self.modules_json.get_modules_json()

        # Ask if we should show the diffs (unless a filename was already given on the command line)
        if not self.save_diff_fn and self.show_diff is None:
            diff_type = questionary.select(
                "Do you want to view diffs of the proposed changes?",
                choices=[
                    {"name": "No previews, just update everything", "value": 0},
                    {"name": "Preview diff in terminal, choose whether to update files", "value": 1},
                    {"name": "Just write diffs to a patch file", "value": 2},
                ],
                default={"name": "No previews, just update everything", "value": 0},
                style=nf_core.utils.nfcore_question_style,
            ).unsafe_ask()

            self.show_diff = diff_type == 1
            self.save_diff_fn = diff_type == 2

        if self.save_diff_fn:  # True or a string
            self.setup_diff_file(check_diff_exist)

        # Loop through all components to be updated
        # and do the requested action on them
        exit_value = True
        all_patches_successful = True
        for modules_repo, component, sha, patch_relpath in components_info:
            if component is None:
                # The entry from .nf-core.yml is set to false, skip update of this component
                continue
            component_fullname = str(Path(self.component_type, modules_repo.repo_path, component))
            # Are we updating the files in place or not?
            dry_run = self.show_diff or self.save_diff_fn

            current_version = self.modules_json.get_component_version(
                self.component_type, component, modules_repo.remote_url, modules_repo.repo_path
            )

            # Set the temporary installation folder
            install_tmp_dir = Path(tempfile.mkdtemp())
            component_install_dir = install_tmp_dir / component

            # Compute the component directory
            component_dir = Path(self.directory, self.component_type, modules_repo.repo_path, component)

            if sha is not None:
                version = sha
            elif self.prompt:
                version = prompt_component_version_sha(
                    component, self.component_type, modules_repo=modules_repo, installed_sha=current_version
                )
            else:
                version = modules_repo.get_latest_component_version(component, self.component_type)

            if current_version is not None and not self.force:
                if current_version == version:
                    if self.sha or self.prompt:
                        log.info(f"'{component_fullname}' is already installed at {version}")
                    else:
                        log.info(f"'{component_fullname}' is already up to date")
                    continue

            # Download component files
            if not self.install_component_files(component, version, modules_repo, str(install_tmp_dir)):
                exit_value = False
                continue

            if patch_relpath is not None:
                patch_successful = self.try_apply_patch(
                    component,
                    modules_repo.repo_path,
                    patch_relpath,
                    component_dir,
                    component_install_dir,
                    write_file=False,
                )
                if patch_successful:
                    log.info(f"{self.component_type[:-1].title()} '{component_fullname}' patched successfully")
                else:
                    log.warning(
                        f"Failed to patch {self.component_type[:-1]} '{component_fullname}'. Will proceed with unpatched files."
                    )
                all_patches_successful &= patch_successful

            if dry_run:
                if patch_relpath is not None:
                    if patch_successful:
                        log.info("Current installation is compared against patched version in remote.")
                    else:
                        log.warning("Current installation is compared against unpatched version in remote.")
                # Compute the diffs for the component
                if self.save_diff_fn:
                    log.info(
                        f"Writing diff file for {self.component_type[:-1]} '{component_fullname}' to '{self.save_diff_fn}'"
                    )
                    try:
                        ComponentsDiffer.write_diff_file(
                            self.save_diff_fn,
                            component,
                            modules_repo.repo_path,
                            component_dir,
                            component_install_dir,
                            current_version,
                            version,
                            dsp_from_dir=component_dir,
                            dsp_to_dir=component_dir,
                            limit_output=self.limit_output,
                        )
                        updated.append(component)
                    except UserWarning as e:
                        if str(e) != "Module is unchanged":
                            raise
                        else:
                            updated.append(component)
                    recursive_update = True
                    modules_to_update, subworkflows_to_update = self.get_components_to_update(component)
                    if not silent and len(modules_to_update + subworkflows_to_update) > 0:
                        log.warning(
                            f"All modules and subworkflows linked to the updated {self.component_type[:-1]} will be added to the same diff file.\n"
                            "It is advised to keep all your modules and subworkflows up to date.\n"
                            "It is not guaranteed that a subworkflow will continue working as expected if all modules/subworkflows used in it are not up to date.\n"
                        )
                        if self.update_deps:
                            recursive_update = True
                        else:
                            recursive_update = questionary.confirm(
                                "Would you like to continue adding all modules and subworkflows differences?",
                                default=True,
                                style=nf_core.utils.nfcore_question_style,
                            ).unsafe_ask()
                    if recursive_update and len(modules_to_update + subworkflows_to_update) > 0:
                        # Write all the differences of linked components to a diff file
                        self.update_linked_components(
                            modules_to_update, subworkflows_to_update, updated, check_diff_exist=False
                        )
                        self.manage_changes_in_linked_components(component, modules_to_update, subworkflows_to_update)

                elif self.show_diff:
                    ComponentsDiffer.print_diff(
                        component,
                        modules_repo.repo_path,
                        component_dir,
                        component_install_dir,
                        current_version,
                        version,
                        dsp_from_dir=component_dir,
                        dsp_to_dir=component_dir,
                        limit_output=self.limit_output,
                    )
                    # Ask the user if they want to install the component
                    dry_run = not questionary.confirm(
                        f"Update {self.component_type[:-1]} '{component}'?",
                        default=False,
                        style=nf_core.utils.nfcore_question_style,
                    ).unsafe_ask()

            if not dry_run:
                # Clear the component directory and move the installed files there
                self.move_files_from_tmp_dir(component, str(install_tmp_dir), modules_repo.repo_path, version)
                # Update modules.json with newly installed component
                self.modules_json.update(self.component_type, modules_repo, component, version, installed_by=None)
                updated.append(component)
                recursive_update = True
                modules_to_update, subworkflows_to_update = self.get_components_to_update(component)
                if not silent and len(modules_to_update + subworkflows_to_update) > 0:
                    if not self.update_all:
                        log.warning(
                            f"All modules and subworkflows linked to the updated {self.component_type[:-1]} will be {'asked for update' if self.show_diff else 'automatically updated'}.\n"
                            "It is advised to keep all your modules and subworkflows up to date.\n"
                            "It is not guaranteed that a subworkflow will continue working as expected if all modules/subworkflows used in it are not up to date.\n"
                        )
                        if self.update_deps:
                            recursive_update = True
                        else:
                            recursive_update = questionary.confirm(
                                "Would you like to continue updating all modules and subworkflows?",
                                default=True,
                                style=nf_core.utils.nfcore_question_style,
                            ).unsafe_ask()
                if recursive_update and len(modules_to_update + subworkflows_to_update) > 0:
                    # Update linked components
                    self.update_linked_components(modules_to_update, subworkflows_to_update, updated)
                    self.manage_changes_in_linked_components(component, modules_to_update, subworkflows_to_update)

        if self.save_diff_fn:
            # Write the modules.json diff to the file
            ComponentsDiffer.append_modules_json_diff(
                self.save_diff_fn,
                old_modules_json,
                self.modules_json.get_modules_json(),
                Path(self.directory, "modules.json"),
            )
            if exit_value and not silent:
                log.info(
                    f"[bold magenta italic] TIP! [/] If you are happy with the changes in '{self.save_diff_fn}', you "
                    "can apply them by running the command :point_right:"
                    f"  [bold magenta italic]git apply {self.save_diff_fn} [/]"
                )
        elif not all_patches_successful and not silent:
            log.info(f"Updates complete. Please apply failed patch{plural_es(components_info)} manually.")
            self.modules_json.load()
            self.modules_json.dump(run_prettier=True)
        elif not silent:
            log.info("Updates complete :sparkles:")
            self.modules_json.load()
            self.modules_json.dump(run_prettier=True)

        return exit_value

    def get_single_component_info(self, component):
        """Collects the modules repository, version and sha for a component.

        Information about the component version in the '.nf-core.yml' overrides
        the '--sha' option

        Args:
            component (str): The name of the module/subworkflow to get info for.

        Returns:
            (ModulesRepo, str, str): The modules repo containing the component,
            the component name, and the component version.

        Raises:
            LookupError: If the component is not found either in the pipeline or the modules repo.
            UserWarning: If the '.nf-core.yml' entry is not valid.
        """
        # Check if there are any modules/subworkflows installed from the repo
        repo_url = self.modules_repo.remote_url
        components = self.modules_json.get_all_components(self.component_type).get(repo_url)
        if components is None:
            raise LookupError(f"No {self.component_type} installed from '{repo_url}'")

        choices = [
            component if directory == self.modules_repo.repo_path else f"{directory}/{component}"
            for directory, component in components
        ]

        if component is None:
            component = questionary.autocomplete(
                f"{self.component_type[:-1].title()} name:",
                choices=sorted(choices),
                style=nf_core.utils.nfcore_question_style,
            ).unsafe_ask()

        # Get component installation directory
        try:
            install_dir = [dir for dir, m in components if component == m][0]
        except IndexError:
            raise UserWarning(f"{self.component_type[:-1].title()} '{component}' not found in 'modules.json'.")

        # Check if component is installed before trying to update
        if component not in choices:
            raise LookupError(
                f"{self.component_type[:-1].title()} '{component}' is not installed in pipeline and could therefore not be updated"
            )

        # Check that the supplied name is an available module/subworkflow
        if component and component not in self.modules_repo.get_avail_components(self.component_type, commit=self.sha):
            raise LookupError(
                f"{self.component_type[:-1].title()} '{component}' not found in list of available {self.component_type}."
                f"Use the command 'nf-core {self.component_type} list remote' to view available software"
            )

        sha = self.sha
        config_entry = None
        if self.update_config is not None:
            if any(
                [
                    entry.count("/") == 1
                    and (entry.endswith("modules") or entry.endswith("subworkflows"))
                    and not (entry.endswith(".git") or entry.endswith(".git/"))
                    for entry in self.update_config.keys()
                ]
            ):
                raise UserWarning(
                    "Your '.nf-core.yml' file format is outdated. "
                    "The format should be of the form:\n"
                    "update:\n  <repo_url>:\n    <component_install_directory>:\n      <component_name>:"
                )
            if isinstance(self.update_config.get(self.modules_repo.remote_url, {}), str):
                # If the repo entry is a string, it's the sha to update to
                config_entry = self.update_config.get(self.modules_repo.remote_url, {})
            elif component in self.update_config.get(self.modules_repo.remote_url, {}).get(install_dir, {}):
                # If the component to update is in .nf-core.yml config file
                config_entry = self.update_config[self.modules_repo.remote_url][install_dir].get(component)
        if config_entry is not None and config_entry is not True:
            if config_entry is False:
                log.warn(
                    f"{self.component_type[:-1].title()}'s update entry in '.nf-core.yml' for '{component}' is set to False"
                )
                return (self.modules_repo, None, None, None)
            if not isinstance(config_entry, str):
                raise UserWarning(
                    f"{self.component_type[:-1].title()}'s update entry in '.nf-core.yml' for '{component}' is of wrong type"
                )

            sha = config_entry
            if self.sha is not None:
                log.warning(
                    f"Found entry in '.nf-core.yml' for {self.component_type[:-1]} '{component}' "
                    "which will override version specified with '--sha'"
                )
            else:
                log.info(f"Found entry in '.nf-core.yml' for {self.component_type[:-1]} '{component}'")
            log.info(f"Updating component to ({sha})")

        # Check if the update branch is the same as the installation branch
        current_branch = self.modules_json.get_component_branch(
            self.component_type, component, self.modules_repo.remote_url, install_dir
        )
        new_branch = self.modules_repo.branch
        if current_branch != new_branch:
            log.warning(
                f"You are trying to update the '{Path(install_dir, component)}' {self.component_type[:-1]} from "
                f"the '{new_branch}' branch. This {self.component_type[:-1]} was installed from the '{current_branch}'"
            )
            switch = questionary.confirm(f"Do you want to update using the '{current_branch}' instead?").unsafe_ask()
            if switch:
                # Change the branch
                self.modules_repo.setup_branch(current_branch)

        # If there is a patch file, get its filename
        patch_fn = self.modules_json.get_patch_fn(
            self.component_type, component, self.modules_repo.remote_url, install_dir
        )

        return (self.modules_repo, component, sha, patch_fn)

    def get_all_components_info(self, branch=None):
        """Collects the modules repository, version and sha for all modules/subworkflows.

        Information about the module/subworkflow version in the '.nf-core.yml' overrides the '--sha' option.

        Returns:
            [(ModulesRepo, str, str)]: A list of tuples containing a ModulesRepo object,
            the component name, and the component version.
        """
        if branch is not None:
            use_branch = questionary.confirm(
                f"'--branch' was specified. Should this branch be used to update all {self.component_type}?",
                default=False,
            )
            if not use_branch:
                branch = None
        skipped_repos = []
        skipped_components = []
        overridden_repos = []
        overridden_components = []
        components_info = {}
        # Loop through all the modules/subworkflows in the pipeline
        # and check if they have an entry in the '.nf-core.yml' file

        for repo_name, components in self.modules_json.get_all_components(self.component_type).items():
            if isinstance(self.update_config, dict) and (
                repo_name not in self.update_config or self.update_config[repo_name] is True
            ):
                # There aren't restrictions for the repository in .nf-core.yml file
                components_info[repo_name] = {}
                for component_dir, component in components:
                    try:
                        components_info[repo_name][component_dir].append(
                            (
                                component,
                                self.sha,
                                self.modules_json.get_component_branch(
                                    self.component_type, component, repo_name, component_dir
                                ),
                            )
                        )
                    except KeyError:
                        components_info[repo_name][component_dir] = [
                            (
                                component,
                                self.sha,
                                self.modules_json.get_component_branch(
                                    self.component_type, component, repo_name, component_dir
                                ),
                            )
                        ]
            elif isinstance(self.update_config, dict) and isinstance(self.update_config[repo_name], dict):
                # If it is a dict, then there are entries for individual components or component directories
                for component_dir in set([dir for dir, _ in components]):
                    if isinstance(self.update_config[repo_name][component_dir], str):
                        # If a string is given it is the commit SHA to which we should update to
                        custom_sha = self.update_config[repo_name][component_dir]
                        components_info[repo_name] = {}
                        for dir, component in components:
                            if component_dir == dir:
                                try:
                                    components_info[repo_name][component_dir].append(
                                        (
                                            component,
                                            custom_sha,
                                            self.modules_json.get_component_branch(
                                                self.component_type, component, repo_name, component_dir
                                            ),
                                        )
                                    )
                                except KeyError:
                                    components_info[repo_name][component_dir] = [
                                        (
                                            component,
                                            custom_sha,
                                            self.modules_json.get_component_branch(
                                                self.component_type, component, repo_name, component_dir
                                            ),
                                        )
                                    ]
                        if self.sha is not None:
                            overridden_repos.append(repo_name)
                    elif self.update_config[repo_name][component_dir] is False:
                        for directory, component in components:
                            if directory == component_dir:
                                skipped_components.append(f"{component_dir}/{components}")
                    elif isinstance(self.update_config[repo_name][component_dir], dict):
                        # If it's a dict, there are entries for individual components
                        dir_config = self.update_config[repo_name][component_dir]
                        components_info[repo_name] = {}
                        for component_dir, component in components:
                            if component not in dir_config or dir_config[component] is True:
                                try:
                                    components_info[repo_name][component_dir].append(
                                        (
                                            component,
                                            self.sha,
                                            self.modules_json.get_component_branch(
                                                self.component_type, component, repo_name, component_dir
                                            ),
                                        )
                                    )
                                except KeyError:
                                    components_info[repo_name][component_dir] = [
                                        (
                                            component,
                                            self.sha,
                                            self.modules_json.get_component_branch(
                                                self.component_type, component, repo_name, component_dir
                                            ),
                                        )
                                    ]
                            elif isinstance(dir_config[component], str):
                                # If a string is given it is the commit SHA to which we should update to
                                custom_sha = dir_config[component]
                                try:
                                    components_info[repo_name][component_dir].append(
                                        (
                                            component,
                                            custom_sha,
                                            self.modules_json.get_component_branch(
                                                self.component_type, component, repo_name, component_dir
                                            ),
                                        )
                                    )
                                except KeyError:
                                    components_info[repo_name][component_dir] = [
                                        (
                                            component,
                                            custom_sha,
                                            self.modules_json.get_component_branch(
                                                self.component_type, component, repo_name, component_dir
                                            ),
                                        )
                                    ]
                                if self.sha is not None:
                                    overridden_components.append(component)
                            elif dir_config[component] is False:
                                # Otherwise the entry must be 'False' and we should ignore the component
                                skipped_components.append(f"{component_dir}/{component}")
                            else:
                                raise UserWarning(
                                    f"{self.component_type[:-1].title()} '{component}' in '{component_dir}' has an invalid entry in '.nf-core.yml'"
                                )
            elif isinstance(self.update_config, dict) and isinstance(self.update_config[repo_name], str):
                # If a string is given it is the commit SHA to which we should update to
                custom_sha = self.update_config[repo_name]
                components_info[repo_name] = {}
                for component_dir, component in components:
                    try:
                        components_info[repo_name][component_dir].append(
                            (
                                component,
                                custom_sha,
                                self.modules_json.get_component_branch(
                                    self.component_type, component, repo_name, component_dir
                                ),
                            )
                        )
                    except KeyError:
                        components_info[repo_name][component_dir] = [
                            (
                                component,
                                custom_sha,
                                self.modules_json.get_component_branch(
                                    self.component_type, component, repo_name, component_dir
                                ),
                            )
                        ]
                if self.sha is not None:
                    overridden_repos.append(repo_name)
            elif isinstance(self.update_config, dict) and self.update_config[repo_name] is False:
                skipped_repos.append(repo_name)
            else:
                log.debug(f"no update config for {repo_name} in `.nf-core.yml`")

        if skipped_repos:
            skipped_str = "', '".join(skipped_repos)
            log.info(f"Skipping {self.component_type} in repositor{plural_y(skipped_repos)}: '{skipped_str}'")

        if skipped_components:
            skipped_str = "', '".join(skipped_components)
            log.info(f"Skipping {self.component_type[:-1]}{plural_s(skipped_components)}: '{skipped_str}'")

        if overridden_repos:
            overridden_str = "', '".join(overridden_repos)
            log.info(
                f"Overriding '--sha' flag for {self.component_type} in repositor{plural_y(overridden_repos)} "
                f"with '.nf-core.yml' entry: '{overridden_str}'"
            )

        if overridden_components:
            overridden_str = "', '".join(overridden_components)
            log.info(
                f"Overriding '--sha' flag for {self.component_type[:-1]}{plural_s(overridden_components)} with "
                f"'.nf-core.yml' entry: '{overridden_str}'"
            )
        # Loop through components_info and create on ModulesRepo object per remote and branch
        repos_and_branches = {}
        for repo_name, repo_content in components_info.items():
            for component_dir, comps in repo_content.items():
                for comp, sha, comp_branch in comps:
                    if branch is not None:
                        comp_branch = branch
                    if (repo_name, comp_branch) not in repos_and_branches:
                        repos_and_branches[(repo_name, comp_branch)] = []
                    repos_and_branches[(repo_name, comp_branch)].append((comp, sha))

        # Create ModulesRepo objects
        repo_objs_comps = []
        for (repo_url, branch), comps_shas in repos_and_branches.items():
            try:
                modules_repo = ModulesRepo(remote_url=repo_url, branch=branch)
            except LookupError as e:
                log.warning(e)
                log.info(f"Skipping {self.component_type} in '{repo_url}'")
            else:
                repo_objs_comps.append((modules_repo, comps_shas))

        # Flatten the list
        components_info = [(repo, comp, sha) for repo, comps_shas in repo_objs_comps for comp, sha in comps_shas]

        # Verify that that all components and shas exist in their respective ModulesRepo,
        # don't try to update those that don't
        i = 0
        while i < len(components_info):
            repo, component, sha = components_info[i]
            if not repo.component_exists(component, self.component_type):
                log.warning(
                    f"{self.component_type[:-1].title()} '{component}' does not exist in '{repo.remote_url}'. Skipping..."
                )
                components_info.pop(i)
            elif sha is not None and not repo.sha_exists_on_branch(sha):
                log.warning(
                    f"Git sha '{sha}' does not exists on the '{repo.branch}' of '{repo.remote_url}'. Skipping {self.component_type[:-1]} '{component}'"
                )
                components_info.pop(i)
            else:
                i += 1

        # Add patch filenames to the components that have them
        components_info = [
            (
                repo,
                comp,
                sha,
                self.modules_json.get_patch_fn(self.component_type, comp, repo.remote_url, repo.repo_path),
            )
            for repo, comp, sha in components_info
        ]

        return components_info

    def setup_diff_file(self, check_diff_exist=True):
        """Sets up the diff file.

        If the save diff option was chosen interactively, the user is asked to supply a name for the diff file.

        Then creates the file for saving the diff.
        """
        if self.save_diff_fn is True:
            # From questionary - no filename yet
            self.save_diff_fn = questionary.path(
                "Enter the filename: ", style=nf_core.utils.nfcore_question_style
            ).unsafe_ask()
        if self.save_diff_fn is not None:
            self.save_diff_fn = Path(self.save_diff_fn)
        else:
            raise UserWarning("No filename provided for saving the diff file")

        if not check_diff_exist:
            # This guarantees that the file exists after calling the function
            self.save_diff_fn.touch()
            return
        # Check if filename already exists (questionary or cli)
        while self.save_diff_fn.exists():
            if questionary.confirm(f"'{self.save_diff_fn}' exists. Remove file?").unsafe_ask():
                os.remove(self.save_diff_fn)
                break
            self.save_diff_fn = questionary.path(
                "Enter a new filename: ",
                style=nf_core.utils.nfcore_question_style,
            ).unsafe_ask()
            self.save_diff_fn = Path(self.save_diff_fn)

        # This guarantees that the file exists after calling the function
        self.save_diff_fn.touch()

    def move_files_from_tmp_dir(self, component: str, install_folder: str, repo_path: str, new_version: str) -> None:
        """Move the files from the temporary to the installation directory.

        Args:
            component (str): The module/subworkflow name.
            install_folder [str]: The path to the temporary installation directory.
            repo_path (str): The name of the directory where modules/subworkflows are installed
            new_version (str): The version of the module/subworkflow that was installed.
        """
        temp_component_dir = Path(install_folder, component)
        files = [file_path for file_path in temp_component_dir.rglob("*") if file_path.is_file()]
        pipeline_path = Path(self.directory, self.component_type, repo_path, component)

        if pipeline_path.exists():
            pipeline_files = [f.name for f in pipeline_path.iterdir() if f.is_file()]
            # check if any *.config file exists in the pipeline
            config_files = [f for f in pipeline_files if str(f).endswith(".config")]
            for config_file in config_files:
                log.debug(f"Moving '{component}/{config_file}' to updated component")
                shutil.move(str(pipeline_path / config_file), temp_component_dir / config_file)
                files.append(temp_component_dir / config_file)

        else:
            log.debug(f"Creating new {self.component_type[:-1]} '{component}' in '{self.component_type}/{repo_path}'")

        log.debug(f"Removing old version of {self.component_type[:-1]} '{component}'")
        self.clear_component_dir(component, str(pipeline_path))

        pipeline_path.mkdir(parents=True, exist_ok=True)
        for file in files:
            file = file.relative_to(temp_component_dir)
            path = Path(temp_component_dir, file)
            if path.exists():
                log.debug(f"Moving '{file}' to updated component")
                dest = Path(pipeline_path, file)
                dest.parent.mkdir(parents=True, exist_ok=True)
                shutil.move(str(path), dest)

        log.info(f"Updating '{repo_path}/{component}'")
        log.debug(f"Updating {self.component_type[:-1]} '{component}' to {new_version} from {repo_path}")

    def try_apply_patch(
        self, component, repo_path, patch_relpath, component_dir, component_install_dir, write_file=True
    ):
        """
        Try applying a patch file to the new module/subworkflow files


        Args:
            component (str): The name of the module/subworkflow
            repo_path (str): The name of the repository where the module/subworkflow resides
            patch_relpath (Path | str): The path to patch file in the pipeline
            component_dir (Path | str): The module/subworkflow directory in the pipeline
            component_install_dir (Path | str): The directory where the new component
                                            file have been installed

        Returns:
            (bool): Whether the patch application was successful
        """
        component_fullname = str(Path(repo_path, component))
        log.info(f"Found patch for  {self.component_type[:-1]} '{component_fullname}'. Trying to apply it to new files")

        patch_path = Path(self.directory / patch_relpath)
        component_relpath = Path(self.component_type, repo_path, component)

        # Check that paths in patch file are updated
        self.check_patch_paths(patch_path, component)

        # Copy the installed files to a new temporary directory to save them for later use
        temp_dir = Path(tempfile.mkdtemp())
        temp_component_dir = temp_dir / component
        shutil.copytree(component_install_dir, temp_component_dir)

        try:
            new_files = ComponentsDiffer.try_apply_patch(
                self.component_type, component, repo_path, patch_path, temp_component_dir
            )
        except LookupError:
            # Patch failed. Save the patch file by moving to the install dir
            shutil.move(patch_path, Path(component_install_dir, patch_path.relative_to(component_dir)))
            log.warning(
                f"Failed to apply patch for {self.component_type[:-1]} '{component_fullname}'. You will have to apply the patch manually"
            )
            return False

        # Write the patched files to a temporary directory
        log.debug("Writing patched files")
        for file, new_content in new_files.items():
            fn = temp_component_dir / file
            with open(fn, "w") as fh:
                fh.writelines(new_content)

        # Create the new patch file
        log.debug("Regenerating patch file")
        ComponentsDiffer.write_diff_file(
            Path(temp_component_dir, patch_path.relative_to(component_dir)),
            component,
            repo_path,
            component_install_dir,
            temp_component_dir,
            file_action="w",
            for_git=False,
            dsp_from_dir=component_relpath,
            dsp_to_dir=component_relpath,
            limit_output=self.limit_output,
        )

        # Move the patched files to the install dir
        log.debug("Overwriting installed files installed files  with patched files")
        shutil.rmtree(component_install_dir)
        shutil.copytree(temp_component_dir, component_install_dir)

        # Add the patch file to the modules.json file
        self.modules_json.add_patch_entry(
            self.component_type,
            component,
            self.modules_repo.remote_url,
            repo_path,
            patch_relpath,
            write_file=write_file,
        )

        return True

    def get_components_to_update(self, component):
        """
        Get all modules and subworkflows linked to the updated component.

        Returns:
            (list,list): A tuple of lists with the modules and subworkflows to update
        """
        mods_json = self.modules_json.get_modules_json()
        modules_to_update = []
        subworkflows_to_update = []
        installed_by = self.modules_json.get_installed_by_entries(self.component_type, component)

        if self.component_type == "modules":
            # All subworkflow names in the installed_by section of a module are subworkflows using this module
            # We need to update them too
            subworkflows_to_update = [subworkflow for subworkflow in installed_by if subworkflow != self.component_type]
        elif self.component_type == "subworkflows":
            for repo, repo_content in mods_json["repos"].items():
                for component_type, dir_content in repo_content.items():
                    for dir, components in dir_content.items():
                        for comp, comp_content in components.items():
                            # If the updated subworkflow name appears in the installed_by section of the checked component
                            # The checked component is used by the updated subworkflow
                            # We need to update it too
                            if component in comp_content["installed_by"]:
                                if component_type == "modules":
                                    modules_to_update.append(comp)
                                elif component_type == "subworkflows":
                                    subworkflows_to_update.append(comp)

        return modules_to_update, subworkflows_to_update

    def update_linked_components(
        self,
        modules_to_update,
        subworkflows_to_update,
        updated=None,
        check_diff_exist=True,
    ):
        """
        Update modules and subworkflows linked to the component being updated.
        """
        for s_update in subworkflows_to_update:
            if s_update in updated:
                continue
            original_component_type, original_update_all = self._change_component_type("subworkflows")
            self.update(
                s_update,
                silent=True,
                updated=updated,
                check_diff_exist=check_diff_exist,
            )
            self._reset_component_type(original_component_type, original_update_all)

        for m_update in modules_to_update:
            if m_update in updated:
                continue
            original_component_type, original_update_all = self._change_component_type("modules")
            try:
                self.update(
                    m_update,
                    silent=True,
                    updated=updated,
                    check_diff_exist=check_diff_exist,
                )
            except LookupError as e:
                # If the module to be updated is not available, check if there has been a name change
                if "not found in list of available" in str(e):
                    # Skip update, we check for name changes with manage_changes_in_linked_components
                    pass
                else:
                    raise
            finally:
                self._reset_component_type(original_component_type, original_update_all)

    def manage_changes_in_linked_components(self, component, modules_to_update, subworkflows_to_update):
        """Check for linked components added or removed in the new subworkflow version"""
        if self.component_type == "subworkflows":
            subworkflow_directory = Path(self.directory, self.component_type, self.modules_repo.repo_path, component)
            included_modules, included_subworkflows = get_components_to_install(subworkflow_directory)
            # If a module/subworkflow has been removed from the subworkflow
            for module in modules_to_update:
                if module not in included_modules:
                    log.info(f"Removing module '{module}' which is not included in '{component}' anymore.")
                    remove_module_object = ComponentRemove("modules", self.directory)
                    remove_module_object.remove(module, removed_by=component)
            for subworkflow in subworkflows_to_update:
                if subworkflow not in included_subworkflows:
                    log.info(f"Removing subworkflow '{subworkflow}' which is not included in '{component}' anymore.")
                    remove_subworkflow_object = ComponentRemove("subworkflows", self.directory)
                    remove_subworkflow_object.remove(subworkflow, removed_by=component)
            # If a new module/subworkflow is included in the subworklfow and wasn't included before
            for module in included_modules:
                if module not in modules_to_update:
                    log.info(f"Installing newly included module '{module}' for '{component}'")
                    install_module_object = ComponentInstall(self.directory, "modules", installed_by=component)
                    install_module_object.install(module, silent=True)
            for subworkflow in included_subworkflows:
                if subworkflow not in subworkflows_to_update:
                    log.info(f"Installing newly included subworkflow '{subworkflow}' for '{component}'")
                    install_subworkflow_object = ComponentInstall(
                        self.directory, "subworkflows", installed_by=component
                    )
                    install_subworkflow_object.install(subworkflow, silent=True)

    def _change_component_type(self, new_component_type):
        original_component_type = self.component_type
        self.component_type = new_component_type
        self.modules_json.pipeline_components = None
        # also reset update_all in case it's set
        original_update_all = self.update_all
        self.update_all = False
        return original_component_type, original_update_all

    def _reset_component_type(self, original_component_type, original_update_all):
        self.component_type = original_component_type
        self.modules_json.pipeline_components = None
        self.update_all = original_update_all



"""
The ComponentCreate class handles generating of module and subworkflow templates
"""

import glob
import json
import logging
import re
import shutil
import subprocess
from pathlib import Path
from typing import Dict, Optional

import jinja2
import questionary
import rich
import rich.prompt
import yaml
from packaging.version import parse as parse_version

import nf_core
import nf_core.utils
from nf_core.components.components_command import ComponentCommand
from nf_core.components.components_utils import get_biotools_id
from nf_core.pipelines.lint_utils import run_prettier_on_file

log = logging.getLogger(__name__)


class ComponentCreate(ComponentCommand):
    def __init__(
        self,
        component_type: str,
        directory: Path = Path("."),
        component: str = "",
        author: Optional[str] = None,
        process_label: Optional[str] = None,
        has_meta: Optional[str] = None,
        force: bool = False,
        conda_name: Optional[str] = None,
        conda_version: Optional[str] = None,
        empty_template: bool = False,
        migrate_pytest: bool = False,
    ):
        super().__init__(component_type, directory)
        self.directory = directory
        self.component = component
        self.author = author
        self.process_label = process_label
        self.has_meta = has_meta
        self.force_overwrite = force
        self.subtool = None
        self.tool_conda_name = conda_name
        self.tool_conda_version = conda_version
        self.tool_licence = ""
        self.tool_description = ""
        self.tool_doc_url = ""
        self.tool_dev_url = ""
        self.bioconda = None
        self.singularity_container = None
        self.docker_container = None
        self.file_paths: Dict[str, Path] = {}
        self.not_empty_template = not empty_template
        self.migrate_pytest = migrate_pytest
        self.tool_identifier = ""

    def create(self) -> bool:
        """
        Create a new DSL2 module or subworkflow from the nf-core template.

        A module should be named just <tool> or <tool/subtool>
        e.g fastqc or samtools/sort, respectively.

        The subworkflow should be named as the main file type it operates on and a short description of the task performed
        e.g bam_sort or bam_sort_samtools, respectively.

        If <directory> is a pipeline, this function creates a file called:
        '<directory>/modules/local/tool.nf'
            OR
        '<directory>/modules/local/tool_subtool.nf'
            OR for subworkflows
        '<directory>/subworkflows/local/subworkflow_name.nf'

        If <directory> is a clone of nf-core/modules, it creates or modifies the following files:

        For modules:

        ```tree
        modules/nf-core/tool/subtool/
        ├── main.nf
        ├── meta.yml
        ├── environment.yml
        └── tests
            └── main.nf.test
        ```

        The function will attempt to automatically find a Bioconda package called <component>
        and matching Docker / Singularity images from BioContainers.

        For subworkflows:

        ```tree
        subworkflows/nf-core/tool/subtool/
        ├── main.nf
        ├── meta.yml
        └── tests
            └── main.nf.test
        ```

        """
        if self.component_type == "modules":
            # Check modules directory structure
            self.check_modules_structure()

        # Check whether the given directory is a nf-core pipeline or a clone of nf-core/modules

        log.info(f"Repository type: [blue]{self.repo_type}")
        if self.directory != ".":
            log.info(f"Base directory: '{self.directory}'")

        log.info(
            "[yellow]Press enter to use default values [cyan bold](shown in brackets)[/] [yellow]or type your own responses. "
            "ctrl+click [link=https://youtu.be/dQw4w9WgXcQ]underlined text[/link] to open links."
        )

        # Collect component info via prompt if empty or invalid
        self._collect_name_prompt()

        # Determine the component name
        self.component_name = self.component
        self.component_dir = Path(self.component)

        if self.subtool:
            self.component_name = f"{self.component}/{self.subtool}"
            self.component_dir = Path(self.component, self.subtool)

        self.component_name_underscore = self.component_name.replace("/", "_")

        # Check existence of directories early for fast-fail
        self.file_paths = self._get_component_dirs()

        if self.migrate_pytest:
            # Rename the component directory to old
            component_old_dir = Path(str(self.component_dir) + "_old")
            component_parent_path = Path(self.directory, self.component_type, self.org)
            component_old_path = component_parent_path / component_old_dir
            component_path = component_parent_path / self.component_dir

            component_path.rename(component_old_path)
        else:
            if self.component_type == "modules":
                # Try to find a bioconda package for 'component'
                self._get_bioconda_tool()
                # Try to find a biotools entry for 'component'
                self.tool_identifier = get_biotools_id(self.component)

            # Prompt for GitHub username
            self._get_username()

            if self.component_type == "modules":
                self._get_module_structure_components()

        # Add a valid organization name for nf-test tags
        not_alphabet = re.compile(r"[^a-zA-Z]")
        self.org_alphabet = not_alphabet.sub("", self.org)

        # Create component template with jinja2
        assert self._render_template()
        log.info(f"Created component template: '{self.component_name}'")

        if self.migrate_pytest:
            self._copy_old_files(component_old_path)
            log.info("Migrate pytest tests: Copied original module files to new module")
            shutil.rmtree(component_old_path)
            self._print_and_delete_pytest_files()

        new_files = [str(path) for path in self.file_paths.values()]

        log.info("Created following files:\n  " + "\n  ".join(new_files))
        return True

    def _get_bioconda_tool(self):
        """
        Try to find a bioconda package for 'tool'
        """
        while True:
            try:
                if self.tool_conda_name:
                    anaconda_response = nf_core.utils.anaconda_package(self.tool_conda_name, ["bioconda"])
                else:
                    anaconda_response = nf_core.utils.anaconda_package(self.component, ["bioconda"])

                if not self.tool_conda_version:
                    version = anaconda_response.get("latest_version")
                    if not version:
                        version = str(max([parse_version(v) for v in anaconda_response["versions"]]))
                else:
                    version = self.tool_conda_version

                self.tool_licence = nf_core.utils.parse_anaconda_licence(anaconda_response, version)
                self.tool_description = anaconda_response.get("summary", "")
                self.tool_doc_url = anaconda_response.get("doc_url", "")
                self.tool_dev_url = anaconda_response.get("dev_url", "")
                if self.tool_conda_name:
                    self.bioconda = "bioconda::" + self.tool_conda_name + "=" + version
                else:
                    self.bioconda = "bioconda::" + self.component + "=" + version
                log.info(f"Using Bioconda package: '{self.bioconda}'")
                break
            except (ValueError, LookupError) as e:
                log.warning(
                    f"Could not find Conda dependency using the Anaconda API: '{self.tool_conda_name if self.tool_conda_name else self.component}'"
                )
                if rich.prompt.Confirm.ask("[violet]Do you want to enter a different Bioconda package name?"):
                    self.tool_conda_name = rich.prompt.Prompt.ask("[violet]Name of Bioconda package").strip()
                    continue
                else:
                    log.warning(
                        f"{e}\nBuilding module without tool software and meta, you will need to enter this information manually."
                    )
                    break

        # Try to get the container tag (only if bioconda package was found)
        if self.bioconda:
            try:
                if self.tool_conda_name:
                    self.docker_container, self.singularity_container = nf_core.utils.get_biocontainer_tag(
                        self.tool_conda_name, version
                    )
                else:
                    self.docker_container, self.singularity_container = nf_core.utils.get_biocontainer_tag(
                        self.component, version
                    )
                log.info(f"Using Docker container: '{self.docker_container}'")
                log.info(f"Using Singularity container: '{self.singularity_container}'")
            except (ValueError, LookupError) as e:
                log.info(f"Could not find a Docker/Singularity container ({e})")

    def _get_module_structure_components(self):
        process_label_defaults = [
            "process_single",
            "process_low",
            "process_medium",
            "process_high",
            "process_long",
            "process_high_memory",
        ]
        if self.process_label is None:
            log.info(
                "Provide an appropriate resource label for the process, taken from the "
                "[link=https://github.com/nf-core/tools/blob/main/nf_core/pipeline-template/conf/base.config#L29]nf-core pipeline template[/link].\n"
                "For example: {}".format(", ".join(process_label_defaults))
            )
        while self.process_label is None:
            self.process_label = questionary.autocomplete(
                "Process resource label:",
                choices=process_label_defaults,
                style=nf_core.utils.nfcore_question_style,
                default="process_single",
            ).unsafe_ask()

        if self.has_meta is None:
            log.info(
                "Where applicable all sample-specific information e.g. 'id', 'single_end', 'read_group' "
                "MUST be provided as an input via a Groovy Map called 'meta'. "
                "This information may [italic]not[/] be required in some instances, for example "
                "[link=https://github.com/nf-core/modules/blob/master/modules/nf-core/bwa/index/main.nf]indexing reference genome files[/link]."
            )
        while self.has_meta is None:
            self.has_meta = rich.prompt.Confirm.ask(
                "[violet]Will the module require a meta map of sample information?",
                default=True,
            )

    def _render_template(self) -> Optional[bool]:
        """
        Create new module/subworkflow files with Jinja2.
        """
        object_attrs = vars(self)
        # Run jinja2 for each file in the template folder
        env = jinja2.Environment(
            loader=jinja2.PackageLoader("nf_core", f"{self.component_type[:-1]}-template"),
            keep_trailing_newline=True,
        )
        for template_fn, dest_fn in self.file_paths.items():
            log.debug(f"Rendering template file: '{template_fn}'")
            j_template = env.get_template(template_fn)
            object_attrs["nf_core_version"] = nf_core.__version__
            try:
                rendered_output = j_template.render(object_attrs)
            except Exception as e:
                log.error(f"Could not render template file '{template_fn}':\n{e}")
                raise e

            # Write output to the target file
            log.debug(f"Writing output to: '{dest_fn}'")
            dest_fn.parent.mkdir(exist_ok=True, parents=True)
            with open(dest_fn, "w") as fh:
                log.debug(f"Writing output to: '{dest_fn}'")
                fh.write(rendered_output)

            # Mirror file permissions
            template_stat = (
                Path(nf_core.__file__).parent / f"{self.component_type[:-1]}-template" / template_fn
            ).stat()
            dest_fn.chmod(template_stat.st_mode)
        return True

    def _collect_name_prompt(self):
        """
        Collect module/subworkflow info via prompt if empty or invalid
        """
        # Collect module info via prompt if empty or invalid
        self.subtool = None
        if self.component_type == "modules":
            pattern = r"[^a-z\d/]"
        elif self.component_type == "subworkflows":
            pattern = r"[^a-z\d_/]"
        if self.component is None:
            self.component = ""
        while self.component == "" or re.search(pattern, self.component) or self.component.count("/") > 0:
            # Check + auto-fix for invalid chacters
            if re.search(pattern, self.component):
                if self.component_type == "modules":
                    log.warning("Tool/subtool name must be lower-case letters only, with no punctuation")
                elif self.component_type == "subworkflows":
                    log.warning("Subworkflow name must be lower-case letters only, with no punctuation")
                name_clean = re.sub(r"[^a-z\d/]", "", self.component.lower())
                if rich.prompt.Confirm.ask(f"[violet]Change '{self.component}' to '{name_clean}'?"):
                    self.component = name_clean
                else:
                    self.component = ""

            if self.component_type == "modules":
                # Split into tool and subtool
                if self.component.count("/") > 1:
                    log.warning("Tool/subtool can have maximum one '/' character")
                    self.component = ""
                elif self.component.count("/") == 1:
                    self.component, self.subtool = self.component.split("/")
                else:
                    self.subtool = None  # Reset edge case: entered '/subtool' as name and gone round loop again

            # Prompt for new entry if we reset
            if self.component == "":
                if self.component_type == "modules":
                    self.component = rich.prompt.Prompt.ask("[violet]Name of tool/subtool").strip()
                elif self.component_type == "subworkflows":
                    self.component = rich.prompt.Prompt.ask("[violet]Name of subworkflow").strip()

    def _get_component_dirs(self) -> Dict[str, Path]:
        """Given a directory and a tool/subtool or subworkflow, set the file paths and check if they already exist

        Returns dict: keys are relative paths to template files, vals are target paths.
        """
        file_paths = {}
        if self.repo_type == "pipeline":
            local_component_dir = Path(self.directory, self.component_type, "local")
            # Check whether component file already exists
            component_file = local_component_dir / f"{self.component_name}.nf"
            if component_file.exists() and not self.force_overwrite:
                raise UserWarning(
                    f"{self.component_type[:-1].title()} file exists already: '{component_file}'. Use '--force' to overwrite"
                )

            if self.component_type == "modules":
                # If a subtool, check if there is a module called the base tool name already
                if self.subtool and (local_component_dir / f"{self.component}.nf").exists():
                    raise UserWarning(
                        f"Module '{self.component}' exists already, cannot make subtool '{self.component_name}'"
                    )

                # If no subtool, check that there isn't already a tool/subtool
                tool_glob = glob.glob(f"{local_component_dir}/{self.component}_*.nf")
                if not self.subtool and tool_glob:
                    raise UserWarning(
                        f"Module subtool '{tool_glob[0]}' exists already, cannot make tool '{self.component_name}'"
                    )

            # Set file paths
            file_paths["main.nf"] = component_file

        elif self.repo_type == "modules":
            component_dir = Path(self.directory, self.component_type, self.org, self.component_dir)
            # Check if module/subworkflow directories exist already
            if component_dir.exists() and not self.force_overwrite and not self.migrate_pytest:
                raise UserWarning(
                    f"{self.component_type[:-1]} directory exists: '{component_dir}'. Use '--force' to overwrite"
                )

            if self.component_type == "modules":
                # If a subtool, check if there is a module called the base tool name already
                parent_tool_main_nf = Path(
                    self.directory,
                    self.component_type,
                    self.org,
                    self.component,
                    "main.nf",
                )
                if self.subtool and parent_tool_main_nf.exists() and not self.migrate_pytest:
                    raise UserWarning(
                        f"Module '{parent_tool_main_nf}' exists already, cannot make subtool '{self.component_name}'"
                    )

                # If no subtool, check that there isn't already a tool/subtool
                tool_glob = glob.glob(
                    f"{Path(self.directory, self.component_type, self.org, self.component)}/*/main.nf"
                )
                if not self.subtool and tool_glob and not self.migrate_pytest:
                    raise UserWarning(
                        f"Module subtool '{tool_glob[0]}' exists already, cannot make tool '{self.component_name}'"
                    )
            # Set file paths
            # For modules - can be tool/ or tool/subtool/ so can't do in template directory structure
            file_paths["main.nf"] = component_dir / "main.nf"
            file_paths["meta.yml"] = component_dir / "meta.yml"
            if self.component_type == "modules":
                file_paths["environment.yml"] = component_dir / "environment.yml"
            file_paths["tests/main.nf.test.j2"] = component_dir / "tests" / "main.nf.test"
        else:
            raise ValueError("`repo_type` not set correctly")

        return file_paths

    def _get_username(self):
        """
        Prompt for GitHub username
        """
        # Try to guess the current user if `gh` is installed
        author_default = None
        try:
            gh_auth_user = json.loads(subprocess.check_output(["gh", "api", "/user"], stderr=subprocess.DEVNULL))
            author_default = f"@{gh_auth_user['login']}"
        except Exception as e:
            log.debug(f"Could not find GitHub username using 'gh' cli command: [red]{e}")

        # Regex to valid GitHub username: https://github.com/shinnn/github-username-regex
        github_username_regex = re.compile(r"^@[a-zA-Z\d](?:[a-zA-Z\d]|-(?=[a-zA-Z\d])){0,38}$")
        while self.author is None or not github_username_regex.match(self.author):
            if self.author is not None and not github_username_regex.match(self.author):
                log.warning("Does not look like a valid GitHub username (must start with an '@')!")
            self.author = rich.prompt.Prompt.ask(
                f"[violet]GitHub Username:[/]{' (@author)' if author_default is None else ''}",
                default=author_default,
            )

    def _copy_old_files(self, component_old_path):
        """Copy files from old module to new module"""
        log.debug("Copying original main.nf file")
        shutil.copyfile(component_old_path / "main.nf", self.file_paths["main.nf"])
        log.debug("Copying original meta.yml file")
        shutil.copyfile(component_old_path / "meta.yml", self.file_paths["meta.yml"])
        if self.component_type == "modules":
            log.debug("Copying original environment.yml file")
            shutil.copyfile(
                component_old_path / "environment.yml",
                self.file_paths["environment.yml"],
            )
            if (component_old_path / "templates").is_dir():
                log.debug("Copying original templates directory")
                shutil.copytree(
                    component_old_path / "templates",
                    self.file_paths["environment.yml"].parent / "templates",
                )
        # Create a nextflow.config file if it contains information other than publishDir
        pytest_dir = Path(self.directory, "tests", self.component_type, self.org, self.component_dir)
        nextflow_config = pytest_dir / "nextflow.config"
        if nextflow_config.is_file():
            with open(nextflow_config) as fh:
                config_lines = ""
                for line in fh:
                    if "publishDir" not in line and line.strip() != "":
                        config_lines += line
            # if the nextflow.config file only contained publishDir, non_publish_dir_lines will be 11 characters long (`process {\n}`)
            if len(config_lines) > 11:
                log.debug("Copying nextflow.config file from pytest tests")
                with open(
                    Path(
                        self.directory,
                        self.component_type,
                        self.org,
                        self.component_dir,
                        "tests",
                        "nextflow.config",
                    ),
                    "w+",
                ) as ofh:
                    ofh.write(config_lines)

    def _print_and_delete_pytest_files(self):
        """Prompt if pytest files should be deleted and printed to stdout"""
        pytest_dir = Path(self.directory, "tests", self.component_type, self.org, self.component_dir)
        if rich.prompt.Confirm.ask(
            "[violet]Do you want to delete the pytest files?[/]\nPytest file 'main.nf' will be printed to standard output to allow migrating the tests manually to 'main.nf.test'.",
            default=False,
        ):
            with open(pytest_dir / "main.nf") as fh:
                log.info(fh.read())
            if pytest_dir.is_symlink():
                resolved_dir = pytest_dir.resolve()
                log.debug(f"Removing symlink: {resolved_dir}")
                shutil.rmtree(resolved_dir)
                pytest_dir.unlink()
            else:
                shutil.rmtree(pytest_dir)
            log.info(
                "[yellow]Please convert the pytest tests to nf-test in 'main.nf.test'.[/]\n"
                "You can find more information about nf-test [link=https://nf-co.re/docs/contributing/modules#migrating-from-pytest-to-nf-test]at the nf-core web[/link]. "
            )
        else:
            log.info(
                "[yellow]Please migrate the pytest tests to nf-test in 'main.nf.test'.[/]\n"
                "You can find more information about nf-test [link=https://nf-co.re/docs/contributing/modules#migrating-from-pytest-to-nf-test]at the nf-core web[/link].\n"
                f"Once done, make sure to delete the module pytest files to avoid linting errors: {pytest_dir}"
            )
        # Delete tags from pytest_modules.yml
        modules_yml = Path(self.directory, "tests", "config", "pytest_modules.yml")
        with open(modules_yml) as fh:
            yml_file = yaml.safe_load(fh)
        yml_key = str(self.component_dir) if self.component_type == "modules" else f"subworkflows/{self.component_dir}"
        if yml_key in yml_file:
            del yml_file[yml_key]
        with open(modules_yml, "w") as fh:
            yaml.dump(yml_file, fh)
        run_prettier_on_file(modules_yml)



from pathlib import Path
from typing import Dict, List, Union

from nf_core.modules.modules_json import ModulesJson, ModulesJsonType


def modules_json(self) -> Dict[str, List[str]]:
    """Make sure all modules described in the ``modules.json`` file are actually installed

    Every module installed from ``nf-core/modules`` must have an entry in the ``modules.json`` file
    with an associated version commit hash.

    * Failure: If module entries are found in ``modules.json`` for modules that are not installed
    """
    passed = []
    warned = []
    failed = []

    # Load pipeline modules and modules.json
    _modules_json = ModulesJson(self.wf_path)
    _modules_json.load()
    modules_json_dict: Union[ModulesJsonType, None] = _modules_json.modules_json
    modules_dir = Path(self.wf_path, "modules")

    if _modules_json and modules_json_dict is not None:
        all_modules_passed = True

        for repo in modules_json_dict["repos"].keys():
            # Check if the modules.json has been updated to keep the
            if "modules" not in modules_json_dict["repos"][repo] or not repo.startswith("http"):
                failed.append(
                    "Your `modules.json` file is outdated. "
                    "It will be automatically generated by running any module command."
                )
                continue

            for dir in modules_json_dict["repos"][repo]["modules"].keys():
                for module, module_entry in modules_json_dict["repos"][repo]["modules"][dir].items():
                    if not Path(modules_dir, dir, module).exists():
                        failed.append(
                            f"Entry for `{Path(modules_dir, dir, module)}` found in `modules.json` but module is not installed in "
                            "pipeline."
                        )
                        all_modules_passed = False
                    if module_entry.get("branch") is None:
                        failed.append(f"Entry for `{Path(modules_dir, dir, module)}` is missing branch information.")
                    if module_entry.get("git_sha") is None:
                        failed.append(f"Entry for `{Path(modules_dir, dir, module)}` is missing version information.")
        if all_modules_passed:
            passed.append("Only installed modules found in `modules.json`")
    else:
        warned.append("Could not open `modules.json` file.")

    return {"passed": passed, "warned": warned, "failed": failed}



import os
import re

import pytest

import nf_core.modules.bump_versions
from nf_core.modules.modules_utils import ModuleExceptionError

from ..test_modules import TestModules


class TestModulesBumpVersions(TestModules):
    def test_modules_bump_versions_single_module(self):
        """Test updating a single module"""
        # Change the bpipe/test version to an older version
        env_yml_path = os.path.join(self.nfcore_modules, "modules", "nf-core", "bpipe", "test", "environment.yml")
        with open(env_yml_path) as fh:
            content = fh.read()
        new_content = re.sub(r"bioconda::star=\d.\d.\d\D?", r"bioconda::star=2.6.1d", content)
        with open(env_yml_path, "w") as fh:
            fh.write(new_content)
        version_bumper = nf_core.modules.bump_versions.ModuleVersionBumper(pipeline_dir=self.nfcore_modules)
        version_bumper.bump_versions(module="bpipe/test")
        assert len(version_bumper.failed) == 0

    def test_modules_bump_versions_all_modules(self):
        """Test updating all modules"""
        version_bumper = nf_core.modules.bump_versions.ModuleVersionBumper(pipeline_dir=self.nfcore_modules)
        version_bumper.bump_versions(all_modules=True)
        assert len(version_bumper.failed) == 0

    def test_modules_bump_versions_fail(self):
        """Fail updating a module with wrong name"""
        version_bumper = nf_core.modules.bump_versions.ModuleVersionBumper(pipeline_dir=self.nfcore_modules)
        with pytest.raises(ModuleExceptionError) as excinfo:
            version_bumper.bump_versions(module="no/module")
        assert "Could not find the specified module:" in str(excinfo.value)

    def test_modules_bump_versions_fail_unknown_version(self):
        """Fail because of an unknown version"""
        # Change the bpipe/test version to an older version
        env_yml_path = os.path.join(self.nfcore_modules, "modules", "nf-core", "bpipe", "test", "environment.yml")
        with open(env_yml_path) as fh:
            content = fh.read()
        new_content = re.sub(r"bioconda::bpipe=\d.\d.\d\D?", r"bioconda::bpipe=xxx", content)
        with open(env_yml_path, "w") as fh:
            fh.write(new_content)
        version_bumper = nf_core.modules.bump_versions.ModuleVersionBumper(pipeline_dir=self.nfcore_modules)
        version_bumper.bump_versions(module="bpipe/test")
        assert "Conda package had unknown version" in version_bumper.failed[0][0]



"""Creates a nf-core pipeline matching the current
organization's specification based on a template.
"""

import configparser
import logging
import os
import re
import shutil
from pathlib import Path
from typing import Dict, List, Optional, Tuple, Union

import git
import git.config
import jinja2
import yaml

import nf_core
import nf_core.pipelines.schema
import nf_core.utils
from nf_core.pipelines.create.utils import CreateConfig, features_yml_path, load_features_yaml
from nf_core.pipelines.create_logo import create_logo
from nf_core.pipelines.lint_utils import run_prettier_on_file
from nf_core.pipelines.rocrate import ROCrate
from nf_core.utils import NFCoreTemplateConfig, NFCoreYamlLintConfig, custom_yaml_dumper

log = logging.getLogger(__name__)


class PipelineCreate:
    """Creates a nf-core pipeline a la carte from the nf-core best-practice template.

    Args:
        name (str): Name for the pipeline.
        description (str): Description for the pipeline.
        author (str): Authors name of the pipeline.
        version (str): Version flag. Semantic versioning only. Defaults to `1.0.0dev`.
        no_git (bool): Prevents the creation of a local Git repository for the pipeline. Defaults to False.
        force (bool): Overwrites a given workflow directory with the same name. Defaults to False. Used for tests and sync command.
            May the force be with you.
        outdir (str): Path to the local output directory.
        template_config (str|CreateConfig): Path to template.yml file for pipeline creation settings. or pydantic model with the customisation for pipeline creation settings.
        organisation (str): Name of the GitHub organisation to create the pipeline. Will be the prefix of the pipeline.
        from_config_file (bool): If true the pipeline will be created from the `.nf-core.yml` config file. Used for tests and sync command.
        default_branch (str): Specifies the --initial-branch name.
    """

    def __init__(
        self,
        name: Optional[str] = None,
        description: Optional[str] = None,
        author: Optional[str] = None,
        version: str = "1.0.0dev",
        no_git: bool = False,
        force: bool = False,
        outdir: Optional[Union[Path, str]] = None,
        template_config: Optional[Union[CreateConfig, str, Path]] = None,
        organisation: str = "nf-core",
        from_config_file: bool = False,
        default_branch: str = "master",
        is_interactive: bool = False,
    ) -> None:
        if isinstance(template_config, CreateConfig):
            self.config = template_config
        elif from_config_file:
            # Try reading config file
            try:
                _, config_yml = nf_core.utils.load_tools_config(outdir if outdir else Path().cwd())
                # Obtain a CreateConfig object from `.nf-core.yml` config file
                if config_yml is not None and getattr(config_yml, "template", None) is not None:
                    self.config = CreateConfig(**config_yml["template"].model_dump(exclude_none=True))
                else:
                    raise UserWarning("The template configuration was not provided in '.nf-core.yml'.")
                # Update the output directory
                self.config.outdir = outdir if outdir else Path().cwd()
            except (FileNotFoundError, UserWarning):
                log.debug("The '.nf-core.yml' configuration file was not found.")
        elif (name and description and author) or (
            template_config and (isinstance(template_config, str) or isinstance(template_config, Path))
        ):
            # Obtain a CreateConfig object from the template yaml file
            self.config = self.check_template_yaml_info(template_config, name, description, author)
            self.update_config(organisation, version, force, outdir)
        else:
            raise UserWarning("The template configuration was not provided.")

        # Read features yaml file
        self.template_features_yml = load_features_yaml()

        # Set fields used by the class methods
        self.no_git = no_git
        self.default_branch = default_branch
        self.is_interactive = is_interactive

        if self.config.outdir is None:
            self.config.outdir = str(Path.cwd())

        # Get the default branch name from the Git configuration
        self.get_default_branch()

        self.jinja_params, self.skip_areas = self.obtain_jinja_params_dict(
            self.config.skip_features or [], str(self.config.outdir)
        )

        # format strings in features yaml
        short_name = self.jinja_params["short_name"]
        env = jinja2.Environment(loader=jinja2.PackageLoader("nf_core", "pipelines"), keep_trailing_newline=True)
        features_template = env.get_template(
            str(features_yml_path.relative_to(Path(nf_core.__file__).parent / "pipelines"))
        )
        rendered_features = features_template.render({"short_name": short_name})
        self.template_features_yml = yaml.safe_load(rendered_features)

        # Get list of files we're skipping with the supplied skip keys
        self.skip_paths = set(sp for k in self.skip_areas for sp in self.template_features_yml[k]["skippable_paths"])

        # Set convenience variables
        self.name = self.config.name
        self.force = self.config.force

        if self.config.outdir == ".":
            self.outdir = Path(self.config.outdir, self.jinja_params["name_noslash"]).absolute()
        else:
            self.outdir = Path(self.config.outdir).absolute()

    def check_template_yaml_info(self, template_yaml, name, description, author):
        """Ensure that the provided template yaml file contains the necessary information.

        Args:
            template_yaml (str): Template yaml file.
            name (str): Name for the pipeline.
            description (str): Description for the pipeline.
            author (str): Authors name of the pipeline.

        Returns:
            CreateConfig: Pydantic model for the nf-core create config.

        Raises:
            UserWarning: if template yaml file does not contain all the necessary information.
            UserWarning: if template yaml file does not exist.
        """
        # Obtain template customization info from template yaml file or `.nf-core.yml` config file
        config = CreateConfig()
        if template_yaml:
            try:
                with open(template_yaml) as f:
                    template_yaml = yaml.safe_load(f)
                    config = CreateConfig(**template_yaml)
            except FileNotFoundError:
                raise UserWarning(f"Template YAML file '{template_yaml}' not found.")

        # Check required fields
        missing_fields = []
        if config.name is None and name is None:
            missing_fields.append("name")
        elif config.name is None:
            config.name = name
        if config.description is None and description is None:
            missing_fields.append("description")
        elif config.description is None:
            config.description = description
        if config.author is None and author is None:
            missing_fields.append("author")
        elif config.author is None:
            config.author = author
        if len(missing_fields) > 0:
            raise UserWarning(
                f"Template YAML file does not contain the following required fields: {', '.join(missing_fields)}"
            )

        return config

    def update_config(self, organisation, version, force, outdir):
        """Updates the config file with arguments provided through command line.

        Args:
            organisation (str): Name of the GitHub organisation to create the pipeline.
            version (str): Version of the pipeline.
            force (bool): Overwrites a given workflow directory with the same name.
            outdir (str): Path to the local output directory.
        """
        if self.config.org is None:
            self.config.org = organisation
        if self.config.version is None:
            self.config.version = version if version else "1.0.0dev"
        if self.config.force is None:
            self.config.force = force if force else False
        if self.config.outdir is None:
            self.config.outdir = outdir if outdir else "."
        if self.config.is_nfcore is None or self.config.is_nfcore == "null":
            self.config.is_nfcore = self.config.org == "nf-core"

    def obtain_jinja_params_dict(
        self, features_to_skip: List[str], pipeline_dir: Union[str, Path]
    ) -> Tuple[Dict, List[str]]:
        """Creates a dictionary of parameters for the new pipeline.

        Args:
            features_to_skip (list<str>): List of template features/areas to skip.
            pipeline_dir (str): Path to the pipeline directory.

        Returns:
            jinja_params (dict): Dictionary of template areas to skip with values true/false.
            skip_areas (list<str>): List of template areas which contain paths to skip.
        """
        # Try reading config file
        try:
            _, config_yml = nf_core.utils.load_tools_config(pipeline_dir)
        except UserWarning:
            config_yml = None

        # Set the parameters for the jinja template
        jinja_params = self.config.model_dump(exclude_none=True)

        # Add template areas to jinja params and create list of areas with paths to skip
        skip_areas = []
        for t_area in self.template_features_yml.keys():
            if t_area in features_to_skip:
                if self.template_features_yml[t_area]["skippable_paths"]:
                    skip_areas.append(t_area)
                jinja_params[t_area] = False
            else:
                jinja_params[t_area] = True

        # Add is_nfcore as an area to skip for non-nf-core pipelines, to skip all nf-core files
        if not self.config.is_nfcore:
            skip_areas.append("is_nfcore")
            jinja_params["is_nfcore"] = False

        # Set the last parameters based on the ones provided
        jinja_params["short_name"] = (
            jinja_params["name"].lower().replace(r"/\s+/", "-").replace(f"{jinja_params['org']}/", "").replace("/", "-")
        )
        jinja_params["name"] = f"{jinja_params['org']}/{jinja_params['short_name']}"
        jinja_params["name_noslash"] = jinja_params["name"].replace("/", "-")
        jinja_params["prefix_nodash"] = jinja_params["org"].replace("-", "")
        jinja_params["name_docker"] = jinja_params["name"].replace(jinja_params["org"], jinja_params["prefix_nodash"])
        jinja_params["logo_light"] = f"{jinja_params['name_noslash']}_logo_light.png"
        jinja_params["logo_dark"] = f"{jinja_params['name_noslash']}_logo_dark.png"
        jinja_params["default_branch"] = self.default_branch
        if config_yml is not None:
            if (
                hasattr(config_yml, "lint")
                and hasattr(config_yml["lint"], "nextflow_config")
                and hasattr(config_yml["lint"]["nextflow_config"], "manifest.name")
            ):
                return jinja_params, skip_areas

        # Check that the pipeline name matches the requirements
        if not re.match(r"^[a-z]+$", jinja_params["short_name"]):
            if jinja_params["is_nfcore"]:
                raise UserWarning("[red]Invalid workflow name: must be lowercase without punctuation.")
            else:
                log.warning(
                    "Your workflow name is not lowercase without punctuation. This may cause Nextflow errors.\nConsider changing the name to avoid special characters."
                )

        return jinja_params, skip_areas

    def init_pipeline(self):
        """Creates the nf-core pipeline."""

        # Make the new pipeline
        self.render_template()

        # Init the git repository and make the first commit
        if not self.no_git:
            self.git_init_pipeline()
            # Run prettier on files
            if self.config.skip_features is None or not (
                "code_linters" in self.config.skip_features or "github" in self.config.skip_features
            ):
                current_dir = Path.cwd()
                os.chdir(self.outdir)
                run_prettier_on_file([str(f) for f in self.outdir.glob("**/*")])
                os.chdir(current_dir)

        if self.config.is_nfcore and not self.is_interactive:
            log.info(
                "[green bold]!!!!!! IMPORTANT !!!!!!\n\n"
                "[green not bold]If you are interested in adding your pipeline to the nf-core community,\n"
                "PLEASE COME AND TALK TO US IN THE NF-CORE SLACK BEFORE WRITING ANY CODE!\n\n"
                "[default]Please read: [link=https://nf-co.re/docs/tutorials/adding_a_pipeline/overview#join-the-community]"
                "https://nf-co.re/docs/tutorials/adding_a_pipeline/overview#join-the-community[/link]"
            )

    def render_template(self) -> None:
        """Runs Jinja to create a new nf-core pipeline."""
        log.info(f"Creating new pipeline: '{self.name}'")

        # Check if the output directory exists
        if self.outdir.exists():
            if self.force:
                log.warning(f"Output directory '{self.outdir}' exists - continuing as --force specified")
            else:
                log.error(f"Output directory '{self.outdir}' exists!")
                log.info("Use -f / --force to overwrite existing files")
                raise UserWarning(f"Output directory '{self.outdir}' exists!")
        else:
            self.outdir.mkdir(parents=True, exist_ok=True)

        # Run jinja2 for each file in the template folder
        env = jinja2.Environment(
            loader=jinja2.PackageLoader("nf_core", "pipeline-template"), keep_trailing_newline=True
        )
        template_dir = Path(nf_core.__file__).parent / "pipeline-template"
        object_attrs = self.jinja_params
        object_attrs["nf_core_version"] = nf_core.__version__
        # Can't use glob.glob() as need recursive hidden dotfiles - https://stackoverflow.com/a/58126417/713980
        template_files = list(Path(template_dir).glob("**/*"))
        template_files += list(Path(template_dir).glob("*"))
        ignore_strs = [".pyc", "__pycache__", ".pyo", ".pyd", ".DS_Store", ".egg"]
        short_name = self.jinja_params["short_name"]
        rename_files: Dict[str, str] = {
            "workflows/pipeline.nf": f"workflows/{short_name}.nf",
            "subworkflows/local/utils_nfcore_pipeline_pipeline/main.nf": f"subworkflows/local/utils_nfcore_{short_name}_pipeline/main.nf",
        }

        # Set the paths to skip according to customization
        for template_fn_path in template_files:
            # Skip files that are in the self.skip_paths list
            for skip_path in self.skip_paths:
                if str(template_fn_path.relative_to(template_dir)).startswith(skip_path):
                    break
            else:
                if template_fn_path.is_dir():
                    continue
                if any([s in str(template_fn_path) for s in ignore_strs]):
                    log.debug(f"Ignoring '{template_fn_path}' in jinja2 template creation")
                    continue

                # Set up vars and directories
                template_fn = template_fn_path.relative_to(template_dir)
                output_path = self.outdir / template_fn

                if str(template_fn) in rename_files:
                    output_path = self.outdir / rename_files[str(template_fn)]
                output_path.parent.mkdir(parents=True, exist_ok=True)

                try:
                    # Just copy binary files
                    if nf_core.utils.is_file_binary(template_fn_path):
                        raise AttributeError(f"Binary file: {template_fn_path}")

                    # Got this far - render the template
                    log.debug(f"Rendering template file: '{template_fn}'")
                    j_template = env.get_template(str(template_fn))
                    rendered_output = j_template.render(object_attrs)

                    # Write to the pipeline output file
                    with open(output_path, "w") as fh:
                        log.debug(f"Writing to output file: '{output_path}'")
                        fh.write(rendered_output)

                # Copy the file directly instead of using Jinja
                except (AttributeError, UnicodeDecodeError) as e:
                    log.debug(f"Copying file without Jinja: '{output_path}' - {e}")
                    shutil.copy(template_fn_path, output_path)

                # Something else went wrong
                except Exception as e:
                    log.error(f"Copying raw file as error rendering with Jinja: '{output_path}' - {e}")
                    shutil.copy(template_fn_path, output_path)

                # Mirror file permissions
                template_stat = os.stat(template_fn_path)
                os.chmod(output_path, template_stat.st_mode)

        if self.config.is_nfcore:
            # Make a logo and save it, if it is a nf-core pipeline
            self.make_pipeline_logo()

        if self.config.skip_features is None or "rocrate" not in self.config.skip_features:
            # Create the RO-Crate metadata file
            rocrate_obj = ROCrate(self.outdir)
            rocrate_obj.create_rocrate(json_path=self.outdir / "ro-crate-metadata.json")

        # Update the .nf-core.yml with linting configurations
        self.fix_linting()

        if self.config:
            config_fn, config_yml = nf_core.utils.load_tools_config(self.outdir)
            if config_fn is not None and config_yml is not None:
                with open(str(config_fn), "w") as fh:
                    config_yml.template = NFCoreTemplateConfig(**self.config.model_dump(exclude_none=True))
                    yaml.dump(config_yml.model_dump(exclude_none=True), fh, Dumper=custom_yaml_dumper())
                    log.debug(f"Dumping pipeline template yml to pipeline config file '{config_fn.name}'")

    def fix_linting(self):
        """
        Updates the .nf-core.yml with linting configurations
        for a customized pipeline.
        """
        # Create a lint config
        lint_config = {}
        for area in (self.config.skip_features or []) + self.skip_areas:
            try:
                for lint_test in self.template_features_yml[area]["linting"]:
                    try:
                        if self.template_features_yml[area]["linting"][lint_test]:
                            lint_config.setdefault(lint_test, []).extend(
                                self.template_features_yml[area]["linting"][lint_test]
                            )
                        else:
                            lint_config[lint_test] = False
                    except AttributeError:
                        pass  # When linting is False
            except KeyError:
                pass  # Areas without linting

        # Add the lint content to the preexisting nf-core config
        config_fn, nf_core_yml = nf_core.utils.load_tools_config(self.outdir)
        if config_fn is not None and nf_core_yml is not None:
            nf_core_yml.lint = NFCoreYamlLintConfig(**lint_config)
            with open(self.outdir / config_fn, "w") as fh:
                yaml.dump(
                    nf_core_yml.model_dump(exclude_none=True),
                    fh,
                    sort_keys=False,
                    default_flow_style=False,
                    Dumper=custom_yaml_dumper(),
                )

    def make_pipeline_logo(self):
        """Fetch a logo for the new pipeline from the nf-core website"""
        email_logo_path = Path(self.outdir) / "assets"
        create_logo(
            text=self.jinja_params["short_name"], directory=email_logo_path, theme="light", force=bool(self.force)
        )
        for theme in ["dark", "light"]:
            readme_logo_path = Path(self.outdir) / "docs" / "images"
            create_logo(
                text=self.jinja_params["short_name"],
                directory=readme_logo_path,
                width=600,
                theme=theme,
                force=bool(self.force),
            )

    def get_default_branch(self) -> None:
        """Gets the default branch name from the Git configuration."""
        try:
            self.default_branch = (
                str(git.config.GitConfigParser().get_value("init", "defaultBranch")) or "master"
            )  # default to master
            log.debug(f"Default branch name: {self.default_branch}")
        except configparser.Error:
            log.debug("Could not read init.defaultBranch")
        if self.default_branch in ["dev", "TEMPLATE"]:
            raise UserWarning(
                f"Your Git defaultBranch '{self.default_branch}' is incompatible with nf-core.\n"
                "'dev' and 'TEMPLATE' can not be used as default branch name.\n"
                "Set the default branch name with "
                "[white on grey23] git config --global init.defaultBranch <NAME> [/]\n"
                "Or set the default_branch parameter in this class.\n"
                "Pipeline git repository will not be initialised."
            )

    def git_init_pipeline(self) -> None:
        """Initialises the new pipeline as a Git repository and submits first commit.

        Raises:
            UserWarning: if Git default branch is set to 'dev' or 'TEMPLATE'.
        """

        log.info("Initialising local pipeline git repository")
        repo = git.Repo.init(self.outdir)
        repo.git.add(A=True)
        repo.index.commit(f"initial template build from nf-core/tools, version {nf_core.__version__}")
        if self.default_branch:
            repo.active_branch.rename(self.default_branch)
        try:
            repo.git.branch("TEMPLATE")
            repo.git.branch("dev")

        except git.GitCommandError as e:
            if "already exists" in e.stderr:
                log.debug("Branches 'TEMPLATE' and 'dev' already exist")
                if self.force:
                    log.debug("Force option set - deleting branches")
                    repo.git.branch("-D", "TEMPLATE")
                    repo.git.branch("-D", "dev")
                    repo.git.branch("TEMPLATE")
                    repo.git.branch("dev")
                else:
                    raise UserWarning(
                        "Branches 'TEMPLATE' and 'dev' already exist. Use --force to overwrite existing branches."
                    )
        if self.is_interactive:
            log.info(f"Pipeline created: ./{self.outdir.relative_to(Path.cwd())}")
        else:
            log.info(
                "Done. Remember to add a remote and push to GitHub:\n"
                f"[white on grey23] cd {self.outdir} \n"
                " git remote add origin git@github.com:USERNAME/REPO_NAME.git \n"
                " git push --all origin                                       "
            )
            log.info("This will also push your newly created dev branch and the TEMPLATE branch for syncing.")



"""Tests covering for utility functions."""

import os
from pathlib import Path
from unittest import mock

import pytest
import requests

import nf_core.pipelines.create.create
import nf_core.pipelines.list
import nf_core.utils

from .test_pipelines import TestPipelines
from .utils import with_temporary_folder

TEST_DATA_DIR = Path(__file__).parent / "data"


def test_strip_ansi_codes():
    """Check that we can make rich text strings plain

    String prints ls examplefile.zip, where examplefile.zip is red bold text
    """
    stripped = nf_core.utils.strip_ansi_codes("ls \x1b[00m\x1b[01;31mexamplefile.zip\x1b[00m\x1b[01;31m")
    assert stripped == "ls examplefile.zip"


class TestUtils(TestPipelines):
    """Class for utils tests"""

    def test_check_if_outdated_1(self):
        current_version = "1.0"
        remote_version = "2.0"
        is_outdated, current, remote = nf_core.utils.check_if_outdated(current_version, remote_version)
        assert is_outdated

    def test_check_if_outdated_2(self):
        current_version = "2.0"
        remote_version = "2.0"
        is_outdated, current, remote = nf_core.utils.check_if_outdated(current_version, remote_version)
        assert not is_outdated

    def test_check_if_outdated_3(self):
        current_version = "2.0.1"
        remote_version = "2.0.2"
        is_outdated, current, remote = nf_core.utils.check_if_outdated(current_version, remote_version)
        assert is_outdated

    def test_check_if_outdated_4(self):
        current_version = "1.10.dev0"
        remote_version = "1.7"
        is_outdated, current, remote = nf_core.utils.check_if_outdated(current_version, remote_version)
        assert not is_outdated

    def test_check_if_outdated_5(self):
        current_version = "1.10.dev0"
        remote_version = "1.11"
        is_outdated, current, remote = nf_core.utils.check_if_outdated(current_version, remote_version)
        assert is_outdated

    def test_rich_force_colours_false(self):
        os.environ.pop("GITHUB_ACTIONS", None)
        os.environ.pop("FORCE_COLOR", None)
        os.environ.pop("PY_COLORS", None)
        assert nf_core.utils.rich_force_colors() is None

    def test_rich_force_colours_true(self):
        os.environ["GITHUB_ACTIONS"] = "1"
        os.environ.pop("FORCE_COLOR", None)
        os.environ.pop("PY_COLORS", None)
        assert nf_core.utils.rich_force_colors() is True

    def testload_pipeline_config(self):
        """Load the pipeline Nextflow config"""
        self.pipeline_obj.load_pipeline_config()
        assert self.pipeline_obj.nf_config["dag.enabled"] == "true"

    # TODO nf-core: Assess and strip out if no longer required for DSL2

    # def test_load_conda_env(self):
    #     """Load the pipeline Conda environment.yml file"""
    #     self.pipeline_obj._load_conda_environment()
    #     assert self.pipeline_obj.conda_config["channels"] == ["conda-forge", "bioconda"]

    def test_list_files_git(self):
        """Test listing pipeline files using `git ls`"""
        files = self.pipeline_obj.list_files()
        assert Path(self.pipeline_dir, "main.nf") in files

    @with_temporary_folder
    def test_list_files_no_git(self, tmpdir):
        """Test listing pipeline files without `git-ls`"""
        # Create a test file in a temporary directory
        tmp_fn = Path(tmpdir, "testfile")
        tmp_fn.touch()
        pipeline_obj = nf_core.utils.Pipeline(tmpdir)
        files = pipeline_obj.list_files()
        assert tmp_fn in files

    @mock.patch("os.path.exists")
    @mock.patch("os.makedirs")
    def test_request_cant_create_cache(self, mock_mkd, mock_exists):
        """Test that we don't get an error when we can't create cachedirs"""
        mock_mkd.side_effect = PermissionError()
        mock_exists.return_value = False
        nf_core.utils.setup_requests_cachedir()

    def test_pip_package_pass(self):
        result = nf_core.utils.pip_package("multiqc=1.10")
        assert isinstance(result, dict)

    @mock.patch("requests.get")
    def test_pip_package_timeout(self, mock_get):
        """Tests the PyPi connection and simulates a request timeout, which should
        return in an additional warning in the linting"""
        # Define the behaviour of the request get mock
        mock_get.side_effect = requests.exceptions.Timeout()
        # Now do the test
        with pytest.raises(LookupError):
            nf_core.utils.pip_package("multiqc=1.10")

    @mock.patch("requests.get")
    def test_pip_package_connection_error(self, mock_get):
        """Tests the PyPi connection and simulates a connection error, which should
        result in an additional warning, as we cannot test if dependent module is latest"""
        # Define the behaviour of the request get mock
        mock_get.side_effect = requests.exceptions.ConnectionError()
        # Now do the test
        with pytest.raises(LookupError):
            nf_core.utils.pip_package("multiqc=1.10")

    def test_pip_erroneous_package(self):
        """Tests the PyPi API package information query"""
        with pytest.raises(ValueError):
            nf_core.utils.pip_package("not_a_package=1.0")

    def test_get_repo_releases_branches_nf_core(self):
        wfs = nf_core.pipelines.list.Workflows()
        wfs.get_remote_workflows()
        pipeline, wf_releases, wf_branches = nf_core.utils.get_repo_releases_branches("methylseq", wfs)
        for r in wf_releases:
            if r.get("tag_name") == "1.6":
                break
        else:
            raise AssertionError("Release 1.6 not found")
        assert "dev" in wf_branches.keys()

    def test_get_repo_releases_branches_not_nf_core(self):
        wfs = nf_core.pipelines.list.Workflows()
        wfs.get_remote_workflows()
        pipeline, wf_releases, wf_branches = nf_core.utils.get_repo_releases_branches("MultiQC/MultiQC", wfs)
        for r in wf_releases:
            if r.get("tag_name") == "v1.10":
                break
        else:
            raise AssertionError("MultiQC release v1.10 not found")
        assert "main" in wf_branches.keys()

    def test_get_repo_releases_branches_not_exists(self):
        wfs = nf_core.pipelines.list.Workflows()
        wfs.get_remote_workflows()
        with pytest.raises(AssertionError):
            nf_core.utils.get_repo_releases_branches("made_up_pipeline", wfs)

    def test_get_repo_releases_branches_not_exists_slash(self):
        wfs = nf_core.pipelines.list.Workflows()
        wfs.get_remote_workflows()
        with pytest.raises(AssertionError):
            nf_core.utils.get_repo_releases_branches("made-up/pipeline", wfs)

    def test_validate_file_md5(self):
        # MD5(test) = d8e8fca2dc0f896fd7cb4cb0031ba249
        test_file = TEST_DATA_DIR / "test.txt"
        test_file_md5 = "d8e8fca2dc0f896fd7cb4cb0031ba249"
        different_md5 = "9e7b964750cf0bb08ee960fce356b6d6"
        non_hex_string = "s"
        assert nf_core.utils.validate_file_md5(test_file, test_file_md5)
        with pytest.raises(IOError):
            nf_core.utils.validate_file_md5(test_file, different_md5)
        with pytest.raises(ValueError):
            nf_core.utils.validate_file_md5(test_file, non_hex_string)

    def test_nested_setitem(self):
        d = {"a": {"b": {"c": "value"}}}
        nf_core.utils.nested_setitem(d, ["a", "b", "c"], "value new")
        assert d["a"]["b"]["c"] == "value new"
        assert d == {"a": {"b": {"c": "value new"}}}

    def test_nested_delitem(self):
        d = {"a": {"b": {"c": "value"}}}
        nf_core.utils.nested_delitem(d, ["a", "b", "c"])
        assert "c" not in d["a"]["b"]
        assert d == {"a": {"b": {}}}

    def test_set_wd(self):
        with nf_core.utils.set_wd(self.tmp_dir):
            context_wd = Path().resolve()
        assert context_wd == Path(self.tmp_dir).resolve()
        assert context_wd != Path().resolve()

    def test_set_wd_revert_on_raise(self):
        wd_before_context = Path().resolve()
        with pytest.raises(Exception):
            with nf_core.utils.set_wd(self.tmp_dir):
                raise Exception
        assert wd_before_context == Path().resolve()



from pathlib import Path

import yaml

import nf_core.pipelines.lint

from ..test_lint import TestLint


class TestLintActionsAws(TestLint):
    def test_actions_awstest_pass(self):
        """Lint test: actions_awstest - PASS"""
        self.lint_obj._load()
        results = self.lint_obj.actions_awstest()
        assert results["passed"] == ["'.github/workflows/awstest.yml' is triggered correctly"]
        assert len(results.get("warned", [])) == 0
        assert len(results.get("failed", [])) == 0
        assert len(results.get("ignored", [])) == 0

    def test_actions_awstest_fail(self):
        """Lint test: actions_awsfulltest - FAIL"""

        # Edit .github/workflows/awsfulltest.yml to use -profile test_full
        new_pipeline = self._make_pipeline_copy()
        with open(Path(new_pipeline, ".github", "workflows", "awstest.yml")) as fh:
            awstest_yml = yaml.safe_load(fh)
        awstest_yml[True]["push"] = ["main"]
        with open(Path(new_pipeline, ".github", "workflows", "awstest.yml"), "w") as fh:
            yaml.dump(awstest_yml, fh)

        # Make lint object
        lint_obj = nf_core.pipelines.lint.PipelineLint(new_pipeline)
        lint_obj._load()

        results = lint_obj.actions_awstest()
        assert results["failed"] == ["'.github/workflows/awstest.yml' is not triggered correctly"]
        assert len(results.get("warned", [])) == 0
        assert len(results.get("passed", [])) == 0
        assert len(results.get("ignored", [])) == 0
